# BSD 2-CLAUSE LICENSE

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:

# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# #ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# original author: Albert Chen
"""Functions to summarize the output of
`~greykite.framework.pipeline.pipeline.forecast_pipeline`.
"""

import re
import warnings

import numpy as np
import pandas as pd
from scipy.stats import rankdata

from greykite.common.constants import FRACTION_OUTSIDE_TOLERANCE
from greykite.common.evaluation import EvaluationMetricEnum
from greykite.common.python_utils import assert_equal
from greykite.framework.constants import CV_REPORT_METRICS_ALL
from greykite.framework.pipeline.utils import get_score_func_with_aggregation


def get_ranks_and_splits(
        grid_search,
        score_func=EvaluationMetricEnum.MeanAbsolutePercentError.name,
        greater_is_better=False,
        combine_splits=True,
        decimals=None,
        warn_metric=True):
    """Extracts CV results from ``grid_search`` for the specified score function.
    Returns the correct ranks on the test set and a tuple of the scores across splits,
    for both test set and train set (if available).

    Notes
    -----
    While ``cv_results`` contains keys with the ranks, these ranks are inverted
    if lower values are better and the ``scoring`` function was initialized
    with ``greater_is_better=True`` to report metrics with their original sign.

    This function always returns the correct ranks, accounting for metric direction.

    Parameters
    ----------
    grid_search : `~sklearn.model_selection.RandomizedSearchCV`
        Grid search output (fitted RandomizedSearchCV object).
    score_func : `str` or callable, default ``EvaluationMetricEnum.MeanAbsolutePercentError.name``
        Score function to get the ranks for.
        If a callable, takes arrays ``y_true``, ``y_pred`` and returns a float.
        If a string, must be either a
        `~greykite.common.evaluation.EvaluationMetricEnum` member name
        or `~greykite.common.constants.FRACTION_OUTSIDE_TOLERANCE`.

        Should be the same as what was passed to
        :py:meth:`~greykite.framework.templates.forecaster.Forecaster.run_forecast_config`,
        or `~greykite.framework.pipeline.pipeline.forecast_pipeline`,
        or `~greykite.framework.pipeline.utils.get_hyperparameter_searcher`.
    greater_is_better : `bool` or None, default False
        True if ``score_func`` is a score function, meaning higher is better,
        and False if it is a loss function, meaning lower is better.
        Must be provided if ``score_func`` is a callable (custom function).
        Ignored if ``score_func`` is a string, because the direction is known.

        Used in this function to rank values in the proper direction.

        Should be the same as what was passed to
        :py:meth:`~greykite.framework.templates.forecaster.Forecaster.run_forecast_config`,
        or `~greykite.framework.pipeline.pipeline.forecast_pipeline`,
        or `~greykite.framework.pipeline.utils.get_hyperparameter_searcher`.
    combine_splits : `bool`, default True
        Whether to report split scores as a tuple in a single column.
        If True, a single column is returned for all the splits
        of a given metric and train/test set.
        For example, "split_train_score" would contain the values
        (split1_train_score, split2_train_score, split3_train_score)
        as as tuple.
        If False, they are reported in their original columns.
    decimals : `int` or None, default None
        Number of decimal places to round to.
        If decimals is negative, it specifies the number of
        positions to the left of the decimal point.
        If None, does not round.
    warn_metric : `bool`, default True
        Whether to issue a warning if the requested metric is
        not found in the CV results.

    Returns
    -------
    ranks_and_splits : `dict`
        Ranks and split scores.
        Dictionary with the following keys:

            ``"short_name"`` : `int`
                Canonical short name for the ``score_func``.
            ``"ranks"`` : `numpy.array`
                Ranks of the test scores for the ``score_func``,
                where 1 is the best.
            ``"split_train"`` : `list` [`list` [`float`]]
                Train split scores. Outer list corresponds to the
                parameter setting; inner list contains the
                scores for that parameter setting across all splits.
            ``"split_test"`` : `list` [`list` [`float`]]
                Test split scores. Outer list corresponds to the
                parameter setting; inner list contains the
                scores for that parameter setting across all splits.
    """
    cv_results = grid_search.cv_results_
    _, greater_is_better, short_name = get_score_func_with_aggregation(
        score_func=score_func,  # string or callable
        greater_is_better=greater_is_better,
        # Dummy value, doesn't matter because we ignore the returned `score_func`
        relative_error_tolerance=0.01)

    # Warns if the metric is not available
    if f"mean_test_{short_name}" not in cv_results:
        if warn_metric:
            warnings.warn(f"Metric '{short_name}' is not available in the CV results.")
        return {
            "short_name": short_name,
            "ranks": None,
            "split_train": None,
            "split_test": None}

    # Computes the ranks, using the same tiebreaking method as in sklearn.
    scores = cv_results[f"mean_test_{short_name}"].copy()
    if greater_is_better:
        scores *= -1  # `rankdata` function ranks lowest values first
    ranks = np.asarray(rankdata(scores, method='min'), dtype=np.int32)

    # Computes split score columns.
    train_scores = None
    test_scores = None

    def round_as_list(split_scores, decimals=None):
        """Rounds `split_scores` to the specified
        `decimals` and returns the result as a list.

        Parameters
        ----------
        split_scores : `numpy.array`
             Split scores.
        decimals : `int` or None, default None
            Number of decimal places to round to.
            If decimals is negative, it specifies the number of
            positions to the left of the decimal point.
            If None, does not round.
        Returns
        -------
        split_scores_list : `list` [`float`]
            ``split_scores``, rounded according
            to ``decimals`` and returned as a list.
        """
        if decimals is not None:
            split_scores = split_scores.round(decimals)
        return split_scores.tolist()

    if combine_splits:
        # Each sublist contains the scores for split i
        # across all parameter settings.
        test_scores = [
            round_as_list(
                cv_results[f"split{i}_test_{short_name}"],
                decimals=decimals)
            for i in range(grid_search.n_splits_)]
        # Makes each sublist contain the scores for a particular
        # parameter setting across all splits.
        test_scores = list(zip(*test_scores))

        # Train scores
        if grid_search.return_train_score:
            train_scores = [
                round_as_list(
                    cv_results[f"split{i}_train_{short_name}"],
                    decimals=decimals)
                for i in range(grid_search.n_splits_)]
            train_scores = list(zip(*train_scores))

    ranks_and_splits = {
        "short_name": short_name,
        "ranks": ranks,
        "split_train": train_scores,
        "split_test": test_scores}
    return ranks_and_splits


def summarize_grid_search_results(
        grid_search,
        only_changing_params=True,
        combine_splits=True,
        decimals=None,
        score_func=EvaluationMetricEnum.MeanAbsolutePercentError.name,
        score_func_greater_is_better=False,
        cv_report_metrics=CV_REPORT_METRICS_ALL,
        column_order=None):
    """Summarizes CV results for each grid search parameter combination.

    While ``grid_search.cv_results_`` could be imported into
    a `pandas.DataFrame` without this function, the following conveniences
    are provided:

        - returns the correct ranks based on each metric's greater_is_better direction.
        - summarizes the hyperparameter space, only showing the parameters that change
        - combines split scores into a tuple to save table width
        - rounds the values to specified decimals
        - orders columns by type (test score, train score, metric, etc.)

    Parameters
    ----------
    grid_search : `~sklearn.model_selection.RandomizedSearchCV`
        Grid search output (fitted RandomizedSearchCV object).
    only_changing_params : `bool`, default True
        If True, only show parameters with multiple values in
        the hyperparameter_grid.
    combine_splits : `bool`, default True
        Whether to report split scores as a tuple in a single column.

            - If True, adds a column for the test splits scores for each
              requested metric. Adds a column with train split scores if those
              are available.

              For example, "split_train_score" would contain the values
              (split1_train_score, split2_train_score, split3_train_score)
              as as tuple.
            - If False, this summary column is not added.

        The original split columns are available either way.
    decimals : `int` or None, default None
        Number of decimal places to round to.
        If decimals is negative, it specifies the number of
        positions to the left of the decimal point.
        If None, does not round.
    score_func : `str` or callable, default ``EvaluationMetricEnum.MeanAbsolutePercentError.name``
        Score function used to select optimal model in CV.
        If a callable, takes arrays ``y_true``, ``y_pred`` and returns a float.
        If a string, must be either a
        `~greykite.common.evaluation.EvaluationMetricEnum` member name
        or `~greykite.common.constants.FRACTION_OUTSIDE_TOLERANCE`.

        Used in this function to fix the ``"rank_test_score"`` column if
        ``score_func_greater_is_better=False``.

        Should be the same as what was passed to
        :py:meth:`~greykite.framework.templates.forecaster.Forecaster.run_forecast_config`,
        or `~greykite.framework.pipeline.pipeline.forecast_pipeline`,
        or `~greykite.framework.pipeline.utils.get_hyperparameter_searcher`.
    score_func_greater_is_better : `bool`, default False
        True if ``score_func`` is a score function, meaning higher is better,
        and False if it is a loss function, meaning lower is better.
        Must be provided if ``score_func`` is a callable (custom function).
        Ignored if ``score_func`` is a string, because the direction is known.

        Used in this function to fix the ``"rank_test_score"`` column if
        ``score_func_greater_is_better=False``.

        Should be the same as what was passed to
        :py:meth:`~greykite.framework.templates.forecaster.Forecaster.run_forecast_config`,
        or `~greykite.framework.pipeline.pipeline.forecast_pipeline`,
        or `~greykite.framework.pipeline.utils.get_hyperparameter_searcher`.
    cv_report_metrics : `~greykite.framework.constants.CV_REPORT_METRICS_ALL`, or `list` [`str`], or None, default `~greykite.common.constants.CV_REPORT_METRICS_ALL`  # noqa: E501
        Additional metrics to show in the summary, besides the one specified by ``score_func``.

        If a metric is specified but not available, a warning will be given.

        Should be the same as what was passed to
        :py:meth:`~greykite.framework.templates.forecaster.Forecaster.run_forecast_config`,
        or `~greykite.framework.pipeline.pipeline.forecast_pipeline`,
        or `~greykite.framework.pipeline.utils.get_hyperparameter_searcher`,
        or a subset of computed metric to show.

        If a list of strings, valid strings are
        `greykite.common.evaluation.EvaluationMetricEnum` member names
        and `~greykite.common.constants.FRACTION_OUTSIDE_TOLERANCE`.
    column_order : `list` [`str`] or None, default None
        How to order the columns.
        A list of regex to order column names, in greedy fashion. Column names matching
        the first item are placed first. Among remaining items, those matching the second
        items are placed next, etc.
        Use "*" as the last element to select all available columns, if desired.
        If None, uses default ordering::

            column_order = ["rank_test", "mean_test", "split_test", "mean_train",
                            "params", "param", "split_train", "time", ".*"]

    Notes
    -----
    Metrics are named in ``grid_search.cv_results_`` according to the ``scoring``
    parameter passed to `~sklearn.model_selection.RandomizedSearchCV`.

    ``"score"`` is the default used by sklearn for single metric
    evaluation.

    If a dictionary is provided to ``scoring``, as is the case through
    templates, then the metrics are named by its keys, and the
    metric used for selection is defined by ``refit``. The keys
    are derived from ``score_func`` and ``cv_report_metrics``
    in `~greykite.framework.pipeline.utils.get_scoring_and_refit`.

        - The key for ``score_func`` if it is a callable is
          `~greykite.common.constants.CUSTOM_SCORE_FUNC_NAME`.
        - The key for ``EvaluationMetricEnum`` member name is the short name
          from ``.get_metric_name()``.
        - The key for `~greykite.common.constants.FRACTION_OUTSIDE_TOLERANCE`
          is `~greykite.common.constants.FRACTION_OUTSIDE_TOLERANCE_NAME`.

    Returns
    -------
    cv_results : `pandas.DataFrame`
        A summary of cross-validation results in tabular format.
        Each row corresponds to a set of parameters used in the grid search.

        The columns have the following format, where name is the canonical short
        name for the metric.

            ``"rank_test_{name}"`` : `int`
                The params ranked by mean_test_score (1 is best).
            ``"mean_test_{name}"`` : `float`
                Average test score.
            ``"split_test_{name}"`` : `list` [`float`]
                Test score on each split. [split 0, split 1, ...]
            ``"std_test_{name}"`` : `float`
                Standard deviation of test scores.
            ``"mean_train_{name}"`` : `float`
                Average train score.
            ``"split_train_{name}"`` : `list` [`float`]
                Train score on each split. [split 0, split 1, ...]
            ``"std_train_{name}"`` : `float`
                Standard deviation of train scores.
            ``"mean_fit_time"`` : `float`
                Average time to fit each CV split (in seconds)
            ``"std_fit_time"`` : `float`
                Std of time to fit each CV split (in seconds)
            ``"mean_score_time"`` : `float`
                Average time to score each CV split (in seconds)
            ``"std_score_time"`` : `float`
                Std of time to score each CV split (in seconds)
            ``"params"`` : `dict`
                The parameters used. If ``only_changing==True``,
                only shows the parameters which are not identical
                across all CV splits.
            ``"param_{pipeline__param__name}"`` : Any
                The value of pipeline parameter `pipeline__param__name`
                for each row.

    """
    if column_order is None:
        column_order = ["rank_test", "mean_test", "split_test", "mean_train", "params", "param", "split_train", "time", ".*"]

    cv_results = grid_search.cv_results_.copy()

    # Overwrites the params
    selected_params = []
    if only_changing_params:
        # Removes keys that don't vary
        keep_params = set()
        seen_params = {}
        for params in cv_results['params']:
            for k, v in params.items():
                if k in seen_params:
                    try:
                        assert_equal(v, seen_params[k])
                    except AssertionError:
                        # the values are different
                        keep_params.add(k)
                else:
                    seen_params[k] = v

        for params in cv_results['params']:
            explore_params = [(k, v) for k, v in params.items() if k in keep_params]
            selected_params.append(explore_params)
        cv_results['params'] = selected_params

    # Overwrites the ranks and computes combined split score columns
    # for the requested metrics.
    metric_list = [(score_func, score_func_greater_is_better, True)]
    if cv_report_metrics == CV_REPORT_METRICS_ALL:
        cv_report_metrics = EvaluationMetricEnum.__dict__["_member_names_"].copy()
        # Computes `FRACTION_OUTSIDE_TOLERANCE` if `relative_error_tolerance` is specified
        cv_report_metrics.append(FRACTION_OUTSIDE_TOLERANCE)
        metric_list += [(metric, None, False) for metric in cv_report_metrics]
    elif cv_report_metrics is not None:
        # greater_is_better is derived from the metric name
        metric_list += [(metric, None, True) for metric in cv_report_metrics]

    keep_metrics = set()
    for metric, greater_is_better, warn_metric in metric_list:
        ranks_and_splits = get_ranks_and_splits(
            grid_search=grid_search,
            score_func=metric,
            greater_is_better=greater_is_better,
            combine_splits=combine_splits,
            decimals=decimals,
            warn_metric=warn_metric)
        short_name = ranks_and_splits["short_name"]
        if ranks_and_splits["ranks"] is not None:
            cv_results[f"rank_test_{short_name}"] = ranks_and_splits["ranks"]
        if ranks_and_splits["split_train"] is not None:
            cv_results[f"split_train_{short_name}"] = ranks_and_splits["split_train"]
        if ranks_and_splits["split_test"] is not None:
            cv_results[f"split_test_{short_name}"] = ranks_and_splits["split_test"]
        keep_metrics.add(short_name)

    # Creates DataFrame and orders the columns.
    # Dictionary keys are unordered, but appears to follow insertion order.
    cv_results_df = pd.DataFrame(cv_results)
    available_cols = list(cv_results_df.columns)
    # Removes metrics not selected
    all_metrics = set(col.replace("mean_test_", "") for col in cv_results.keys()
                      if re.search("mean_test_", col))
    remove_metrics = all_metrics - keep_metrics
    remove_regex = "|".join(remove_metrics)
    if remove_regex:
        available_cols = [col for col in available_cols
                          if not re.search(remove_regex, col)]
    # Orders the columns
    ordered_cols = []
    for regex in column_order:
        selected_cols = [col for col in available_cols
                         if col not in ordered_cols and re.search(regex, col)]
        ordered_cols += selected_cols
    cv_results_df = cv_results_df[ordered_cols]

    if decimals is not None:
        cv_results_df = cv_results_df.round(decimals)

    return cv_results_df



# BSD 2-CLAUSE LICENSE

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:

# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# #ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# original author: Albert Chen

import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator
from sklearn.base import TransformerMixin
from sklearn.exceptions import NotFittedError

from greykite.common.logging import LoggingLevelEnum
from greykite.common.logging import log_message


class ZscoreOutlierTransformer(BaseEstimator, TransformerMixin):
    """Replaces outliers in data with NaN.
    Outliers are determined by z-score cutoff. Columns are handled independently.

    Parameters
    ----------
    z_cutoff : `float` or None, default None
        z-score cutoff to define outliers. If None, this transformer is a no-op.
    use_fit_baseline : `bool`, default False
        If True, the z-scores are calculated using the mean and standard
        deviation of the dataset passed to ``fit``.

        If False, the transformer is stateless. z-scores are calculated
        for the dataset passed to ``transform``, regardless of ``fit``.

    Attributes
    ----------
    mean : `pandas.Series`
        Mean of each column. NaNs are ignored.
    std : `pandas.Series`
        Standard deviation of each column. NaNs are ignored.
    _is_fitted : `bool`
        Whether the transformer is fitted.
    """
    def __init__(self, z_cutoff=None, use_fit_baseline=False):
        # sets params without modification to ensure get_params() works in grid search
        self.z_cutoff = z_cutoff
        self.use_fit_baseline = use_fit_baseline

        self.mean = None
        self.std = None
        self._is_fitted = None

    def fit(self, X, y=None):
        """Computes the column mean and standard deviation,
        stored as ``mean`` and ``std`` attributes.

        Parameters
        ----------
        X : `pandas.DataFrame`
            Training input data. e.g. each column is a timeseries.
            Columns are expected to be numeric.
        y : None
            There is no need of a target in a transformer, yet the pipeline API
            requires this parameter.

        Returns
        -------
        self : object
            Returns self.
        """
        assert isinstance(X, pd.DataFrame)
        self._is_fitted = True
        if self.z_cutoff is not None and self.use_fit_baseline:
            self.mean = X.mean()
            self.std = X.std()
        return self

    def transform(self, X):
        """Replaces outliers with NaN.

        Parameters
        ----------
        X : `pandas.DataFrame`
            Data to transform. e.g. each column is a timeseries.
            Columns are expected to be numeric.

        Returns
        -------
        X_outlier : `pandas.DataFrame`
            A copy of the data frame with original values and outliers replaced with NaN.
        """
        assert isinstance(X, pd.DataFrame)
        result = X.copy()
        if self.z_cutoff is not None:
            if self.use_fit_baseline:
                if self._is_fitted is None:
                    raise NotFittedError(
                        "This instance is not fitted yet. Call 'fit' with appropriate arguments "
                        "before calling 'transform'.")
                mean = self.mean
                std = self.std
            else:
                mean = X.mean()
                std = X.std()
            outlier_indices = np.abs(X - mean) > std * self.z_cutoff
            if np.any(outlier_indices):
                total_na = outlier_indices.sum().sum()
                log_message(f"Detected {total_na} outlier(s).", LoggingLevelEnum.INFO)
            result = result.mask(outlier_indices)
        return result



# BSD 2-CLAUSE LICENSE

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:

# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# #ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# original author: Albert Chen
"""Provides default parameter values for ``ForecastConfig``, shared across templates.
The "*_template.py" files contain defaults particular to a specific template.
"""
import dataclasses
from typing import List
from typing import Optional
from typing import Union

from greykite.common.constants import TIME_COL
from greykite.common.constants import VALUE_COL
from greykite.common.evaluation import EvaluationMetricEnum
from greykite.framework.constants import COMPUTATION_N_JOBS
from greykite.framework.constants import COMPUTATION_VERBOSE
from greykite.framework.constants import CV_REPORT_METRICS_ALL
from greykite.framework.constants import EVALUATION_PERIOD_CV_MAX_SPLITS
from greykite.framework.templates.autogen.forecast_config import ComputationParam
from greykite.framework.templates.autogen.forecast_config import EvaluationMetricParam
from greykite.framework.templates.autogen.forecast_config import EvaluationPeriodParam
from greykite.framework.templates.autogen.forecast_config import ForecastConfig
from greykite.framework.templates.autogen.forecast_config import MetadataParam
from greykite.framework.templates.autogen.forecast_config import ModelComponentsParam


class ForecastConfigDefaults:
    """Class that applies default values to a
    `~greykite.framework.templates.autogen.forecast_config.ForecastConfig` object.

    Provides these methods:

        - apply_metadata_defaults
        - apply_evaluation_metric_defaults
        - apply_evaluation_period_defaults
        - apply_computation_defaults
        - apply_model_components_defaults
        - apply_forecast_config_defaults

    Subclasses may override these if different defaults are desired.
    """
    DEFAULT_MODEL_TEMPLATE = "SILVERKITE"
    """The default model template. See `~greykite.framework.templates.model_templates.ModelTemplateEnum`.
    Uses a string to avoid circular imports.
    """

    @staticmethod
    def apply_computation_defaults(computation: Optional[ComputationParam] = None) -> ComputationParam:
        """Applies the default ComputationParam values to the given object.
        If an expected attribute value is provided, the value is unchanged. Otherwise the default value for it is used.
        Other attributes are untouched.
        If the input object is None, it creates a ComputationParam object.

        Parameters
        ----------
        computation : `~greykite.framework.templates.autogen.forecast_config.ComputationParam` or None
            The ComputationParam object.

        Returns
        -------
        computation : `~greykite.framework.templates.autogen.forecast_config.ComputationParam`
            Valid ComputationParam object with the provided attribute values and the default attribute values if not.
        """
        if computation is None:
            computation = ComputationParam()
        if computation.n_jobs is None:
            computation.n_jobs = COMPUTATION_N_JOBS
        if computation.verbose is None:
            computation.verbose = COMPUTATION_VERBOSE
        return computation

    @staticmethod
    def apply_evaluation_metric_defaults(evaluation: Optional[EvaluationMetricParam] = None) -> EvaluationMetricParam:
        """Applies the default EvaluationMetricParam values to the given object.
        If an expected attribute value is provided, the value is unchanged. Otherwise the default value for it is used.
        Other attributes are untouched.
        If the input object is None, it creates a EvaluationMetricParam object.

        Parameters
        ----------
        evaluation : `~greykite.framework.templates.autogen.forecast_config.EvaluationMetricParam` or None
            The EvaluationMetricParam object.

        Returns
        -------
        evaluation : `~greykite.framework.templates.autogen.forecast_config.EvaluationMetricParam`
            Valid EvaluationMetricParam object with the provided attribute values and the default attribute values if not.
        """
        if evaluation is None:
            evaluation = EvaluationMetricParam()
        if evaluation.cv_selection_metric is None:
            # NB: subclass may want to override, if designed for a different objective (e.g. quantile loss)
            evaluation.cv_selection_metric = EvaluationMetricEnum.MeanAbsolutePercentError.name
        if evaluation.cv_report_metrics is None:
            evaluation.cv_report_metrics = CV_REPORT_METRICS_ALL
        return evaluation

    @staticmethod
    def apply_evaluation_period_defaults(evaluation: Optional[EvaluationPeriodParam] = None) -> EvaluationPeriodParam:
        """Applies the default EvaluationPeriodParam values to the given object.
        If an expected attribute value is provided, the value is unchanged. Otherwise the default value for it is used.
        Other attributes are untouched.
        If the input object is None, it creates a EvaluationPeriodParam object.

        Parameters
        ----------
        evaluation : `~greykite.framework.templates.autogen.forecast_config.EvaluationPeriodParam` or None
            The EvaluationMetricParam object.

        Returns
        -------
        evaluation : `~greykite.framework.templates.autogen.forecast_config.EvaluationPeriodParam`
            Valid EvaluationPeriodParam object with the provided attribute values and the default attribute values if not.
        """
        if evaluation is None:
            evaluation = EvaluationPeriodParam()
        if evaluation.cv_max_splits is None:
            evaluation.cv_max_splits = EVALUATION_PERIOD_CV_MAX_SPLITS
        if evaluation.cv_periods_between_train_test is None:
            evaluation.cv_periods_between_train_test = evaluation.periods_between_train_test
        if evaluation.cv_expanding_window is None:
            # NB: subclass may want to override.
            evaluation.cv_expanding_window = True   # good for long-term forecasts, or when data are limited
        return evaluation

    @staticmethod
    def apply_metadata_defaults(metadata: Optional[MetadataParam] = None) -> MetadataParam:
        """Applies the default MetadataParam values to the given object.
        If an expected attribute value is provided, the value is unchanged. Otherwise the default value for it is used.
        Other attributes are untouched.
        If the input object is None, it creates a MetadataParam object.

        Parameters
        ----------
        metadata : `~greykite.framework.templates.autogen.forecast_config.MetadataParam` or None
            The MetadataParam object.

        Returns
        -------
        metadata : `~greykite.framework.templates.autogen.forecast_config.MetadataParam`
            Valid MetadataParam object with the provided attribute values and the default attribute values if not.
        """
        if metadata is None:
            metadata = MetadataParam()
        if metadata.time_col is None:
            metadata.time_col = TIME_COL
        if metadata.value_col is None:
            metadata.value_col = VALUE_COL
        return metadata

    @staticmethod
    def apply_model_components_defaults(model_components: Optional[Union[ModelComponentsParam, List[Optional[ModelComponentsParam]]]] = None) \
            -> Union[ModelComponentsParam, List[ModelComponentsParam]]:
        """Applies the default ModelComponentsParam values to the given object.

        Converts None to a ModelComponentsParam object.
        Unpacks a list of a single element to the element itself.

        Parameters
        ----------
        model_components : `~greykite.framework.templates.autogen.forecast_config.ModelComponentsParam` or None or list of such items
            The ModelComponentsParam object.

        Returns
        -------
        model_components : `~greykite.framework.templates.autogen.forecast_config.ModelComponentsParam` or list of such items
            Valid ModelComponentsParam object with the provided attribute values and the default attribute values if not.
        """
        # Converts single element to a list
        if not isinstance(model_components, list):
            model_components = [model_components]
        # Replaces all `None` with ModelComponentsParam()
        model_components = [m if m is not None else ModelComponentsParam() for m in model_components]
        # model_components can be provided as a list or a single element.
        # A list of a single element is unpacked to that element.
        # (Some template classes like SilverkiteTemplate do not allow model_components
        # to be a list.)
        if isinstance(model_components, list) and len(model_components) == 1:
            model_components = model_components[0]
        return model_components

    def apply_model_template_defaults(self, model_template: Optional[Union[str, List[Optional[str]]]] = None) -> Union[str, List[str]]:
        """Applies the default ModelComponentsParam values to the given object.

        Unpacks a list of a single element to the element itself.
        Sets default value if None.

        Parameters
        ----------
        model_template : `str` or None or `list` [None, `str`]
            The model template name.
            See valid names in `~greykite.framework.templates.model_templates.ModelTemplateEnum`.

        Returns
        -------
        model_template : `str` or `list` [`str`]
            The model template name, with defaults value used if not provided.
        """
        # Converts single element to a list
        if not isinstance(model_template, list):
            model_template = [model_template]
        model_template = [m if m is not None else self.DEFAULT_MODEL_TEMPLATE for m in model_template]
        if isinstance(model_template, list) and len(model_template) == 1:
            # model_template can be provided as a list or a single element.
            # A list of a single element is unpacked to that element.
            # (Some template classes like SilverkiteTemplate do not allow model_template
            # to be a list.)
            model_template = model_template[0]
        return model_template

    def apply_forecast_config_defaults(self, config: Optional[ForecastConfig] = None) -> ForecastConfig:
        """Applies the default Forecast Config values to the given config.
        If an expected attribute value is provided, the value is unchanged.
        Otherwise the default value for it is used.
        Other attributes are untouched.
        If the input config is None, it creates a Forecast Config.

        Parameters
        ----------
        config : :class:`~greykite.framework.templates.autogen.forecast_config.ForecastConfig` or None
            Forecast configuration if available. See
            :class:`~greykite.framework.templates.autogen.forecast_config.ForecastConfig`.

        Returns
        -------
        config : :class:`~greykite.framework.templates.model_templates.ForecastConfig`
            A valid Forecast Config which contains the provided attribute values and the default attribute values if not.
        """
        if config is None:
            config = ForecastConfig()
        else:
            # makes a copy to avoid mutating input
            config = dataclasses.replace(config)

        config.computation_param = self.apply_computation_defaults(config.computation_param)
        config.evaluation_metric_param = self.apply_evaluation_metric_defaults(config.evaluation_metric_param)
        config.evaluation_period_param = self.apply_evaluation_period_defaults(config.evaluation_period_param)
        config.metadata_param = self.apply_metadata_defaults(config.metadata_param)
        config.model_components_param = self.apply_model_components_defaults(config.model_components_param)
        config.model_template = self.apply_model_template_defaults(config.model_template)
        return config



# BSD 2-CLAUSE LICENSE

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:

# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# #ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# original author: Kaixu Yang
"""Provides templates for SimpleSilverkiteEstimator that are pre-tuned to fit
specific use cases.

A subset of these templates are recognized by
`~greykite.framework.templates.ModelTemplateEnum`.

`~greykite.framework.templates.simple_silverkite_template`
also accepts any ``model_template`` name that follows
the naming convention in this file. For details, see
the ``model_template`` parameter in
`~greykite.framework.templates.simple_silverkite_template.SimpleSilverkiteTemplate`.
"""

from dataclasses import dataclass
from enum import Enum
from typing import Dict
from typing import List
from typing import Type
from typing import Union

from greykite.common.python_utils import mutable_field
from greykite.framework.templates.autogen.forecast_config import ModelComponentsParam


# Defines the keywords used in default templates.
# The first item in each value is the default value.
# None of the keys can be any of the keywords, otherwise the algorithm will fail to parse the template name.
class SILVERKITE_FREQ(Enum):
    """Valid values for simple silverkite template string name frequency."""
    HOURLY = "HOURLY"
    DAILY = "DAILY"
    WEEKLY = "WEEKLY"
    DEFAULT = "DAILY"


VALID_FREQ = [name.value for name in SILVERKITE_FREQ if name.name != "DEFAULT"]
"""Valid non-default values for simple silverkite template string name frequency.
These are the non-default frequencies recognized by
`~greykite.framework.templates.simple_silverkite_template_config.SimpleSilverkiteTemplateOptions`.
"""


class SILVERKITE_SEAS(Enum):
    """Valid values for simple silverkite template string name seasonality.
    """
    LT = "LT"
    NM = "NM"
    HV = "HV"
    LTQM = "LTQM"
    NMQM = "NMQM"
    HVQM = "HVQM"
    NONE = "NONE"
    DEFAULT = "LT"


class SILVERKITE_GR(Enum):
    """Valid values for simple silverkite template string name growth_term.
    """
    LINEAR = "LINEAR"
    NONE = "NONE"
    DEFAULT = "LINEAR"


class SILVERKITE_CP(Enum):
    """Valid values for simple silverkite template string name changepoints_dict.
    """
    NONE = "NONE"
    LT = "LT"
    NM = "NM"
    HV = "HV"
    DEFAULT = "NONE"


class SILVERKITE_HOL(Enum):
    """Valid values for simple silverkite template string name events.
    """
    NONE = "NONE"
    SP1 = "SP1"
    SP2 = "SP2"
    SP4 = "SP4"
    TG = "TG"
    DEFAULT = "NONE"


class SILVERKITE_FEASET(Enum):
    """Valid values for simple silverkite template string name feature_sets_enabled.
    """
    OFF = "OFF"
    AUTO = "AUTO"
    ON = "ON"
    DEFAULT = "OFF"


class SILVERKITE_ALGO(Enum):
    """Valid values for simple silverkite template string name fit_algorithm.
    """
    LINEAR = "LINEAR"
    RIDGE = "RIDGE"
    SGD = "SGD"
    LASSO = "LASSO"
    DEFAULT = "LINEAR"


class SILVERKITE_AR(Enum):
    """Valid values for simple silverkite template string name autoregression.
    """
    OFF = "OFF"
    AUTO = "AUTO"
    DEFAULT = "OFF"


class SILVERKITE_DSI(Enum):
    """Valid values for simple silverkite template string name daily seasonality max interaction order.
    """
    AUTO = "AUTO"
    OFF = "OFF"
    DEFAULT = "AUTO"


class SILVERKITE_WSI(Enum):
    """Valid values for simple silverkite template string name weekly seasonality max interaction order.
    """
    AUTO = "AUTO"
    OFF = "OFF"
    DEFAULT = "AUTO"


# Defines the enum that has all keywords.
class SILVERKITE_COMPONENT_KEYWORDS(Enum):
    """Valid values for simple silverkite template string name keywords.
    The names are the keywords and the values are the corresponding value enum.
    Can be used to create an instance of
    `~greykite.framework.templates.simple_silverkite_template_config.SimpleSilverkiteTemplateOptions`.
    """
    FREQ = SILVERKITE_FREQ
    SEAS = SILVERKITE_SEAS
    GR = SILVERKITE_GR
    CP = SILVERKITE_CP
    HOL = SILVERKITE_HOL
    FEASET = SILVERKITE_FEASET
    ALGO = SILVERKITE_ALGO
    AR = SILVERKITE_AR
    DSI = SILVERKITE_DSI
    WSI = SILVERKITE_WSI


# Defines the data class of all keywords.
@dataclass
class SimpleSilverkiteTemplateOptions:
    """Defines generic simple silverkite template options.

    Attributes can be set to different values using
    `~greykite.framework.templates.simple_silverkite_template_config.SILVERKITE_COMPONENT_KEYWORDS`
    for high level tuning.

    ``freq`` represents data frequency.

    The other attributes stand for seasonality,
    growth, changepoints_dict, events, feature_sets_enabled, fit_algorithm and autoregression in
    `~greykite.framework.templates.autogen.forecast_config.ModelComponentsParam`, which are used in
    `~greykite.framework.templates.simple_silverkite_template.SimpleSilverkiteTemplate`.
    """
    freq: SILVERKITE_FREQ = SILVERKITE_FREQ.DEFAULT
    """Valid values for simple silverkite template string name frequency.
    See `~greykite.framework.templates.simple_silverkite_template_config.SILVERKITE_FREQ`.
    """
    seas: SILVERKITE_SEAS = SILVERKITE_SEAS.DEFAULT
    """Valid values for simple silverkite template string name seasonality.
    See `~greykite.framework.templates.simple_silverkite_template_config.SILVERKITE_SEAS`.
    """
    gr: SILVERKITE_GR = SILVERKITE_GR.DEFAULT
    """Valid values for simple silverkite template string name growth.
    See `~greykite.framework.templates.simple_silverkite_template_config.SILVERKITE_GR`.
    """
    cp: SILVERKITE_CP = SILVERKITE_CP.DEFAULT
    """Valid values for simple silverkite template string name changepoints.
    See `~greykite.framework.templates.simple_silverkite_template_config.SILVERKITE_CP`.
    """
    hol: SILVERKITE_HOL = SILVERKITE_HOL.DEFAULT
    """Valid values for simple silverkite template string name holiday.
    See `~greykite.framework.templates.simple_silverkite_template_config.SILVERKITE_HOL`.
    """
    feaset: SILVERKITE_FEASET = SILVERKITE_FEASET.DEFAULT
    """Valid values for simple silverkite template string name feature sets enabled.
    See `~greykite.framework.templates.simple_silverkite_template_config.SILVERKITE_FEASET`.
    """
    algo: SILVERKITE_ALGO = SILVERKITE_ALGO.DEFAULT
    """Valid values for simple silverkite template string name fit algorithm.
    See `~greykite.framework.templates.simple_silverkite_template_config.SILVERKITE_ALGO`.
    """
    ar: SILVERKITE_AR = SILVERKITE_AR.DEFAULT
    """Valid values for simple silverkite template string name autoregression.
    See `~greykite.framework.templates.simple_silverkite_template_config.SILVERKITE_AR`.
    """
    dsi: SILVERKITE_DSI = SILVERKITE_DSI.DEFAULT
    """Valid values for simple silverkite template string name max daily seasonality interaction order.
    See `~greykite.framework.templates.simple_silverkite_template_config.SILVERKITE_DSI`.
    """
    wsi: SILVERKITE_WSI = SILVERKITE_WSI.DEFAULT
    """Valid values for simple silverkite template string name max weekly seasonality interaction order.
    See `~greykite.framework.templates.simple_silverkite_template_config.SILVERKITE_WSI`.
    """


# Defines the common single parameters in `ModelComponentsParam`.
COMMON_MODELCOMPONENTPARAM_PARAMETERS = dict(
    # Seasonality components and orders depending on frequency HOURLY/DAILY/WEEKLY.
    SEAS=dict(
        HOURLY=dict(
            LT={
                "yearly_seasonality": 8,
                "quarterly_seasonality": 0,
                "monthly_seasonality": 0,
                "weekly_seasonality": 3,
                "daily_seasonality": 5
            },
            NM={
                "yearly_seasonality": 15,
                "quarterly_seasonality": 0,
                "monthly_seasonality": 0,
                "weekly_seasonality": 4,
                "daily_seasonality": 8
            },
            HV={
                "yearly_seasonality": 25,
                "quarterly_seasonality": 0,
                "monthly_seasonality": 0,
                "weekly_seasonality": 6,
                "daily_seasonality": 12
            },
            LTQM={
                "yearly_seasonality": 8,
                "quarterly_seasonality": 2,
                "monthly_seasonality": 2,
                "weekly_seasonality": 3,
                "daily_seasonality": 5
            },
            NMQM={
                "yearly_seasonality": 15,
                "quarterly_seasonality": 3,
                "monthly_seasonality": 3,
                "weekly_seasonality": 4,
                "daily_seasonality": 8
            },
            HVQM={
                "yearly_seasonality": 25,
                "quarterly_seasonality": 4,
                "monthly_seasonality": 4,
                "weekly_seasonality": 6,
                "daily_seasonality": 12
            },
            NONE={
                "yearly_seasonality": 0,
                "quarterly_seasonality": 0,
                "monthly_seasonality": 0,
                "weekly_seasonality": 0,
                "daily_seasonality": 0
            }
        ),
        DAILY=dict(
            LT={
                "yearly_seasonality": 8,
                "quarterly_seasonality": 0,
                "monthly_seasonality": 0,
                "weekly_seasonality": 3,
                "daily_seasonality": 0
            },
            NM={
                "yearly_seasonality": 15,
                "quarterly_seasonality": 0,
                "monthly_seasonality": 0,
                "weekly_seasonality": 3,
                "daily_seasonality": 0
            },
            HV={
                "yearly_seasonality": 25,
                "quarterly_seasonality": 0,
                "monthly_seasonality": 0,
                "weekly_seasonality": 4,
                "daily_seasonality": 0
            },
            LTQM={
                "yearly_seasonality": 8,
                "quarterly_seasonality": 3,
                "monthly_seasonality": 2,
                "weekly_seasonality": 3,
                "daily_seasonality": 0
            },
            NMQM={
                "yearly_seasonality": 15,
                "quarterly_seasonality": 4,
                "monthly_seasonality": 4,
                "weekly_seasonality": 3,
                "daily_seasonality": 0
            },
            HVQM={
                "yearly_seasonality": 25,
                "quarterly_seasonality": 6,
                "monthly_seasonality": 4,
                "weekly_seasonality": 4,
                "daily_seasonality": 0
            },
            NONE={
                "yearly_seasonality": 0,
                "quarterly_seasonality": 0,
                "monthly_seasonality": 0,
                "weekly_seasonality": 0,
                "daily_seasonality": 0
            }
        ),
        WEEKLY=dict(
            LT={
                "yearly_seasonality": 8,
                "quarterly_seasonality": 0,
                "monthly_seasonality": 0,
                "weekly_seasonality": 0,
                "daily_seasonality": 0
            },
            NM={
                "yearly_seasonality": 15,
                "quarterly_seasonality": 0,
                "monthly_seasonality": 0,
                "weekly_seasonality": 0,
                "daily_seasonality": 0
            },
            HV={
                "yearly_seasonality": 25,
                "quarterly_seasonality": 0,
                "monthly_seasonality": 0,
                "weekly_seasonality": 0,
                "daily_seasonality": 0
            },
            LTQM={
                "yearly_seasonality": 8,
                "quarterly_seasonality": 2,
                "monthly_seasonality": 2,
                "weekly_seasonality": 0,
                "daily_seasonality": 0
            },
            NMQM={
                "yearly_seasonality": 15,
                "quarterly_seasonality": 3,
                "monthly_seasonality": 3,
                "weekly_seasonality": 0,
                "daily_seasonality": 0
            },
            HVQM={
                "yearly_seasonality": 25,
                "quarterly_seasonality": 4,
                "monthly_seasonality": 4,
                "weekly_seasonality": 0,
                "daily_seasonality": 0
            },
            NONE={
                "yearly_seasonality": 0,
                "quarterly_seasonality": 0,
                "monthly_seasonality": 0,
                "weekly_seasonality": 0,
                "daily_seasonality": 0
            }
        )),
    # Growth function.
    GR=dict(
        LINEAR={
            "growth_term": "linear"
        },
        NONE={
            "growth_term": None
        }),
    # Trend changepoints depending on frequency HOURLY/DAILY/WEEKLY.
    CP=dict(
        HOURLY=dict(
            LT={
                "method": "auto",
                "resample_freq": "D",
                "regularization_strength": 0.6,
                "potential_changepoint_distance": "7D",
                "no_changepoint_distance_from_end": "30D",
                "yearly_seasonality_order": 15,
                "yearly_seasonality_change_freq": None
            },
            NM={
                "method": "auto",
                "resample_freq": "D",
                "regularization_strength": 0.5,
                "potential_changepoint_distance": "15D",
                "no_changepoint_distance_from_end": "30D",
                "yearly_seasonality_order": 15,
                "yearly_seasonality_change_freq": "365D"
            },
            HV={
                "method": "auto",
                "resample_freq": "D",
                "regularization_strength": 0.3,
                "potential_changepoint_distance": "15D",
                "no_changepoint_distance_from_end": "30D",
                "yearly_seasonality_order": 15,
                "yearly_seasonality_change_freq": "365D"
            },
            NONE=None
        ),
        DAILY=dict(
            LT={
                "method": "auto",
                "resample_freq": "7D",
                "regularization_strength": 0.6,
                "potential_changepoint_distance": "15D",
                "no_changepoint_distance_from_end": "90D",
                "yearly_seasonality_order": 15,
                "yearly_seasonality_change_freq": None
            },
            NM={
                "method": "auto",
                "resample_freq": "7D",
                "regularization_strength": 0.5,
                "potential_changepoint_distance": "15D",
                "no_changepoint_distance_from_end": "180D",
                "yearly_seasonality_order": 15,
                "yearly_seasonality_change_freq": "365D"
            },
            HV={
                "method": "auto",
                "resample_freq": "7D",
                "regularization_strength": 0.3,
                "potential_changepoint_distance": "15D",
                "no_changepoint_distance_from_end": "180D",
                "yearly_seasonality_order": 15,
                "yearly_seasonality_change_freq": "365D"
            },
            NONE=None
        ),
        WEEKLY=dict(
            LT={
                "method": "auto",
                "resample_freq": "7D",
                "regularization_strength": 0.6,
                "potential_changepoint_distance": "14D",
                "no_changepoint_distance_from_end": "180D",
                "yearly_seasonality_order": 15,
                "yearly_seasonality_change_freq": None
            },
            NM={
                "method": "auto",
                "resample_freq": "7D",
                "regularization_strength": 0.5,
                "potential_changepoint_distance": "14D",
                "no_changepoint_distance_from_end": "180D",
                "yearly_seasonality_order": 15,
                "yearly_seasonality_change_freq": "365D"
            },
            HV={
                "method": "auto",
                "resample_freq": "7D",
                "regularization_strength": 0.3,
                "potential_changepoint_distance": "14D",
                "no_changepoint_distance_from_end": "180D",
                "yearly_seasonality_order": 15,
                "yearly_seasonality_change_freq": "365D"
            },
            NONE=None
        )),
    # Holiday effect.
    HOL=dict(
        SP1={
            "holidays_to_model_separately": "auto",
            "holiday_lookup_countries": "auto",
            "holiday_pre_num_days": 1,
            "holiday_post_num_days": 1,
            "holiday_pre_post_num_dict": None,
            "daily_event_df_dict": None,
        },
        SP2={
            "holidays_to_model_separately": "auto",
            "holiday_lookup_countries": "auto",
            "holiday_pre_num_days": 2,
            "holiday_post_num_days": 2,
            "holiday_pre_post_num_dict": None,
            "daily_event_df_dict": None,
        },
        SP4={
            "holidays_to_model_separately": "auto",
            "holiday_lookup_countries": "auto",
            "holiday_pre_num_days": 4,
            "holiday_post_num_days": 4,
            "holiday_pre_post_num_dict": None,
            "daily_event_df_dict": None,
        },
        TG={
            "holidays_to_model_separately": [],
            "holiday_lookup_countries": "auto",
            "holiday_pre_num_days": 3,
            "holiday_post_num_days": 3,
            "holiday_pre_post_num_dict": None,
            "daily_event_df_dict": None,
        },
        NONE={
            "holidays_to_model_separately": [],
            "holiday_lookup_countries": [],
            "holiday_pre_num_days": 0,
            "holiday_post_num_days": 0,
            "holiday_pre_post_num_dict": None,
            "daily_event_df_dict": None,
        }),
    # Feature sets enabled.
    FEASET=dict(
        AUTO="auto",
        ON=True,
        OFF=False),
    # Fit algorithm.
    ALGO=dict(
        LINEAR={
            "fit_algorithm": "linear",
            "fit_algorithm_params": None
        },
        RIDGE={
            "fit_algorithm": "ridge",
            "fit_algorithm_params": None
        },
        SGD={
            "fit_algorithm": "sgd",
            "fit_algorithm_params": None
        },
        LASSO={
            "fit_algorithm": "lasso",
            "fit_algorithm_params": None
        }),
    # Autoregression.
    AR=dict(
        AUTO={
            "autoreg_dict": "auto"
        },
        OFF={
            "autoreg_dict": None
        }),
    # Max daily/weekly seasonality interaction orders.
    DSI=dict(
        HOURLY=dict(
            AUTO=5,
            OFF=0
        ),
        DAILY=dict(
            AUTO=0,
            OFF=0
        ),
        WEEKLY=dict(
            AUTO=0,
            OFF=0
        )
    ),
    WSI=dict(
        HOURLY=dict(
            AUTO=2,
            OFF=0
        ),
        DAILY=dict(
            AUTO=2,
            OFF=0
        ),
        WEEKLY=dict(
            AUTO=0,
            OFF=0
        )
    )
)
"""Defines the default component values for
`~greykite.framework.templates.simple_silverkite_template.SimpleSilverkiteTemplate`.
The components include seasonality, growth, holiday, trend changepoints,
feature sets, autoregression, fit algorithm, etc. These are used when
`config.model_template` provides the
`~greykite.framework.templates.simple_silverkite_template_config.SimpleSilverkiteTemplateOptions`.
"""


# Defines the SILVERKITE template here.
SILVERKITE = ModelComponentsParam(
    seasonality={
        "yearly_seasonality": "auto",
        "quarterly_seasonality": "auto",
        "monthly_seasonality": "auto",
        "weekly_seasonality": "auto",
        "daily_seasonality": "auto",
    },
    growth={
        "growth_term": "linear"
    },
    events={
        "holidays_to_model_separately": "auto",
        "holiday_lookup_countries": "auto",
        "holiday_pre_num_days": 2,
        "holiday_post_num_days": 2,
        "holiday_pre_post_num_dict": None,
        "daily_event_df_dict": None,
    },
    changepoints={
        "changepoints_dict": None,
        "seasonality_changepoints_dict": None
    },
    autoregression={
        "autoreg_dict": None
    },
    regressors={
        "regressor_cols": []
    },
    uncertainty={
        "uncertainty_dict": None
    },
    custom={
        "fit_algorithm_dict": {
            "fit_algorithm": "ridge",
            "fit_algorithm_params": None,
        },
        "feature_sets_enabled": "auto",  # "auto" based on data freq and size
        "max_daily_seas_interaction_order": 5,
        "max_weekly_seas_interaction_order": 2,
        "extra_pred_cols": [],
        "min_admissible_value": None,
        "max_admissible_value": None,
    }
)
"""Defines the ``SILVERKITE`` template. Contains automatic growth,
seasonality, holidays, and interactions. Does not include autoregression.
Best for hourly and daily frequencies. Uses `SimpleSilverkiteEstimator`.
"""

# Defines the `SILVERKITE_EMPTY` template. Everything here is None or off.
# The "DAILY" here does not make any difference from "HOURLY" or "WEEKLY" when everything is None or off.
SILVERKITE_EMPTY = "DAILY_SEAS_NONE_GR_NONE_CP_NONE_HOL_NONE_FEASET_OFF_ALGO_LINEAR_AR_OFF_DSI_OFF_WSI_OFF"


# Defines pre-tuned multi-templates.
SILVERKITE_DAILY_90 = [
    # For daily data, light seasonality up to weekly, light trend changepoints,
    # separate holidays +- 2 days, default feature sets and linear fit algorithm.
    "DAILY_SEAS_LTQM_GR_LINEAR_CP_LT_HOL_SP2_FEASET_AUTO_ALGO_LINEAR_AR_OFF_DSI_AUTO_WSI_AUTO",
    # For daily data, light seasonality up to weekly, no trend changepoints,
    # separate holidays +- 2 days, default feature sets and linear fit algorithm.
    "DAILY_SEAS_LTQM_GR_LINEAR_CP_NONE_HOL_SP2_FEASET_AUTO_ALGO_LINEAR_AR_OFF_DSI_AUTO_WSI_AUTO",
    # For daily data, light seasonality up to weekly, light trend changepoints,
    # separate holidays +- 2 days, default feature sets and ridge fit algorithm.
    "DAILY_SEAS_LTQM_GR_LINEAR_CP_LT_HOL_SP2_FEASET_AUTO_ALGO_RIDGE_AR_OFF_DSI_AUTO_WSI_AUTO",
    # For daily data, normal seasonality up to weekly, light trend changepoints,
    # separate holidays +- 4 days, default feature sets and ridge fit algorithm.
    "DAILY_SEAS_NM_GR_LINEAR_CP_LT_HOL_SP4_FEASET_AUTO_ALGO_RIDGE_AR_OFF_DSI_AUTO_WSI_AUTO"
]
SILVERKITE_WEEKLY = [
    # For weekly data, normal seasonality up to yearly, no trend changepoints,
    # no holiday, no feature sets and linear fit algorithm.
    "WEEKLY_SEAS_NM_GR_LINEAR_CP_NONE_HOL_NONE_FEASET_OFF_ALGO_LINEAR_AR_OFF_DSI_AUTO_WSI_AUTO",
    # For weekly data, normal seasonality up to yearly, light trend changepoints,
    # no holiday, no feature sets and linear fit algorithm.
    "WEEKLY_SEAS_NM_GR_LINEAR_CP_LT_HOL_NONE_FEASET_OFF_ALGO_LINEAR_AR_OFF_DSI_AUTO_WSI_AUTO",
    # For weekly data, heavy seasonality up to yearly, normal trend changepoints,
    # no holiday, no feature sets and ridge fit algorithm.
    "WEEKLY_SEAS_HV_GR_LINEAR_CP_NM_HOL_NONE_FEASET_OFF_ALGO_RIDGE_AR_OFF_DSI_AUTO_WSI_AUTO",
    # For weekly data, heavy seasonality up to yearly, light trend changepoints,
    # no holiday, no feature sets and ridge fit algorithm.
    "WEEKLY_SEAS_HV_GR_LINEAR_CP_LT_HOL_NONE_FEASET_OFF_ALGO_RIDGE_AR_OFF_DSI_AUTO_WSI_AUTO"
]
SILVERKITE_HOURLY_1 = [
    # For hourly data, light seasonality up to daily, no trend changepoints,
    # together holiday, default feature sets, automatic autoregression and linear fit algorithm.
    "HOURLY_SEAS_LT_GR_LINEAR_CP_NONE_HOL_TG_FEASET_AUTO_ALGO_LINEAR_AR_AUTO",
    # For hourly data, normal seasonality up to daily, light trend changepoints,
    # separate holidays +- 4 days, default feature sets, automatic autoregression and linear fit algorithm.
    "HOURLY_SEAS_NM_GR_LINEAR_CP_LT_HOL_SP4_FEASET_AUTO_ALGO_LINEAR_AR_AUTO",
    # For hourly data, light seasonality up to daily, normal trend changepoints,
    # separate holidays +- 4 days, no feature sets, automatic autoregression and ridge fit algorithm.
    "HOURLY_SEAS_LT_GR_LINEAR_CP_NM_HOL_SP4_FEASET_OFF_ALGO_RIDGE_AR_AUTO",
    # For hourly data, normal seasonality up to daily, normal trend changepoints,
    # separate holidays +- 1 day, default feature sets, automatic autoregression and ridge fit algorithm.
    "HOURLY_SEAS_NM_GR_LINEAR_CP_NM_HOL_SP1_FEASET_AUTO_ALGO_RIDGE_AR_AUTO"
]
SILVERKITE_HOURLY_24 = [
    # For hourly data, light seasonality up to daily, normal trend changepoints,
    # separate holidays +- 4 days, default feature sets, automatic autoregression and ridge fit algorithm.
    "HOURLY_SEAS_LT_GR_LINEAR_CP_NM_HOL_SP4_FEASET_AUTO_ALGO_RIDGE_AR_AUTO",
    # For hourly data, light seasonality up to daily, no trend changepoints,
    # separate holidays +- 4 days, default feature sets, automatic autoregression and ridge fit algorithm.
    "HOURLY_SEAS_LT_GR_LINEAR_CP_NONE_HOL_SP4_FEASET_AUTO_ALGO_RIDGE_AR_AUTO",
    # For hourly data, normal seasonality up to daily, light trend changepoints,
    # separate holidays +- 1 days, no feature sets, automatic autoregression and linear fit algorithm.
    "HOURLY_SEAS_NM_GR_LINEAR_CP_LT_HOL_SP1_FEASET_OFF_ALGO_LINEAR_AR_AUTO",
    # For hourly data, normal seasonality up to daily, normal trend changepoints,
    # separate holidays +- 4 day, default feature sets, automatic autoregression and ridge fit algorithm.
    "HOURLY_SEAS_NM_GR_LINEAR_CP_NM_HOL_SP4_FEASET_AUTO_ALGO_RIDGE_AR_AUTO"
]
SILVERKITE_HOURLY_168 = [
    # For hourly data, light seasonality up to daily, light trend changepoints,
    # separate holidays +- 4 days, default feature sets, no autoregression and ridge fit algorithm.
    "HOURLY_SEAS_LT_GR_LINEAR_CP_LT_HOL_SP4_FEASET_AUTO_ALGO_RIDGE_AR_OFF",
    # For hourly data, light seasonality up to daily, light trend changepoints,
    # separate holidays +- 2 days, default feature sets, no autoregression and ridge fit algorithm.
    "HOURLY_SEAS_LT_GR_LINEAR_CP_LT_HOL_SP2_FEASET_AUTO_ALGO_RIDGE_AR_OFF",
    # For hourly data, normal seasonality up to daily, no trend changepoints,
    # separate holidays +- 4 days, no feature sets, automatic autoregression and linear fit algorithm.
    "HOURLY_SEAS_NM_GR_LINEAR_CP_NONE_HOL_SP4_FEASET_OFF_ALGO_LINEAR_AR_AUTO",
    # For hourly data, normal seasonality up to daily, normal trend changepoints,
    # separate holidays +- 1 day, default feature sets, no autoregression and ridge fit algorithm.
    "HOURLY_SEAS_NM_GR_LINEAR_CP_NM_HOL_SP1_FEASET_AUTO_ALGO_RIDGE_AR_OFF"
]
SILVERKITE_HOURLY_336 = [
    # For hourly data, light seasonality up to daily, light trend changepoints,
    # separate holidays +- 2 days, default feature sets, no autoregression and ridge fit algorithm.
    "HOURLY_SEAS_LT_GR_LINEAR_CP_LT_HOL_SP2_FEASET_AUTO_ALGO_RIDGE_AR_OFF",
    # For hourly data, light seasonality up to daily, light trend changepoints,
    # separate holidays +- 4 days, default feature sets, no autoregression and ridge fit algorithm.
    "HOURLY_SEAS_LT_GR_LINEAR_CP_LT_HOL_SP4_FEASET_AUTO_ALGO_RIDGE_AR_OFF",
    # For hourly data, normal seasonality up to daily, light trend changepoints,
    # separate holidays +- 2 days, default feature sets, no autoregression and linear fit algorithm.
    "HOURLY_SEAS_NM_GR_LINEAR_CP_LT_HOL_SP1_FEASET_AUTO_ALGO_LINEAR_AR_OFF",
    # For hourly data, normal seasonality up to daily, normal trend changepoints,
    # separate holidays +- 1 day, default feature sets, automatic autoregression and linear fit algorithm.
    "HOURLY_SEAS_NM_GR_LINEAR_CP_NM_HOL_SP1_FEASET_AUTO_ALGO_LINEAR_AR_AUTO"
]


# Defines pre-defined multi templates.

MULTI_TEMPLATES = {
    "SILVERKITE_DAILY_90": SILVERKITE_DAILY_90,
    "SILVERKITE_WEEKLY": SILVERKITE_WEEKLY,
    "SILVERKITE_HOURLY_1": SILVERKITE_HOURLY_1,
    "SILVERKITE_HOURLY_24": SILVERKITE_HOURLY_24,
    "SILVERKITE_HOURLY_168": SILVERKITE_HOURLY_168,
    "SILVERKITE_HOURLY_336": SILVERKITE_HOURLY_336
}
"""A dictionary of multi templates.

    - Keys are the available multi templates names (valid strings for `config.model_template`).
    - Values correspond to a list of
      `~greykite.framework.templates.autogen.forecast_config.ModelComponentsParam`.
"""

SINGLE_MODEL_TEMPLATE_TYPE = Union[str, ModelComponentsParam, SimpleSilverkiteTemplateOptions]
"""Types accepted by SimpleSilverkiteTemplate for ``config.model_template``
for a single template.
"""


@dataclass
class SimpleSilverkiteTemplateConstants:
    """Constants used by
    `~greykite.framework.templates.simple_silverkite_template.SimpleSilverkiteTemplate`.
    Includes the model templates and their default values.

    `~greykite.common.python_utils.mutable_field` is used when the default value
    is a mutable type like dict and list. Dataclass requires mutable default values
    to be wrapped in 'default_factory', so that instances of this dataclass cannot
    accidentally modify the default value.
    `~greykite.common.python_utils.mutable_field` wraps the constant accordingly.
    """
    COMMON_MODELCOMPONENTPARAM_PARAMETERS: Dict = mutable_field(COMMON_MODELCOMPONENTPARAM_PARAMETERS)
    """Defines the default component values for
    `~greykite.framework.templates.simple_silverkite_template.SimpleSilverkiteTemplate`.
    The components include seasonality, growth, holiday, trend changepoints,
    feature sets, autoregression, fit algorithm, etc. These are used when
    `config.model_template` provides the
    `~greykite.framework.templates.simple_silverkite_template_config.SimpleSilverkiteTemplateOptions`.
    """
    MULTI_TEMPLATES: Dict = mutable_field(MULTI_TEMPLATES)
    """A dictionary of multi templates.

        - Keys are the available multi templates names (valid strings for `config.model_template`).
        - Values correspond to a list of
          `~greykite.framework.templates.autogen.forecast_config.ModelComponentsParam`.
    """
    SILVERKITE: SINGLE_MODEL_TEMPLATE_TYPE = SILVERKITE
    """Defines the ``"SILVERKITE"`` template. Contains automatic growth,
    seasonality, holidays, and interactions. Does not include autoregression.
    Best for hourly and daily frequencies. Uses `SimpleSilverkiteEstimator`.
    """
    SILVERKITE_COMPONENT_KEYWORDS: Type[Enum] = SILVERKITE_COMPONENT_KEYWORDS
    """Valid values for simple silverkite template string name keywords.
    The names are the keywords and the values are the corresponding value enum.
    Can be used to create an instance of
    `~greykite.framework.templates.simple_silverkite_template_config.SimpleSilverkiteTemplateOptions`.
    """
    SILVERKITE_EMPTY: SINGLE_MODEL_TEMPLATE_TYPE = SILVERKITE_EMPTY
    """Defines the ``"SILVERKITE_EMPTY"`` template. Everything here is None or off."""
    VALID_FREQ: List = mutable_field(VALID_FREQ)
    """Valid non-default values for simple silverkite template string name frequency.
    `~greykite.framework.templates.simple_silverkite_template_config.SimpleSilverkiteTemplateOptions`.
    """
    SimpleSilverkiteTemplateOptions: dataclass = SimpleSilverkiteTemplateOptions
    """Defines generic simple silverkite template options.
    Attributes can be set to different values using
    `~greykite.framework.templates.simple_silverkite_template_config.SILVERKITE_COMPONENT_KEYWORDS`
    for high level tuning.
    """



# BSD 2-CLAUSE LICENSE

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:

# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# #ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# original author: Reza Hosseini
"""Preprocessing function to handle anomalies in input data
before forecasting.
"""

import warnings

import numpy as np
import pandas as pd

from greykite.common.constants import END_DATE_COL
from greykite.common.constants import START_DATE_COL


def adjust_anomalous_data(
        df,
        time_col,
        value_col,
        anomaly_df,
        start_date_col=START_DATE_COL,
        end_date_col=END_DATE_COL,
        adjustment_delta_col=None,
        filter_by_dict=None,
        filter_by_value_col=None,
        adjustment_method="add"):
    """This function takes:

        - a time series, in the form of a dataframe: ``df``
        - the anomaly information, in the form of a dataframe: ``anomaly_df``.

    It then adjusts the values of the time series based on the perceived impact
    of the anomalies given in the column ``adjustment_delta_col`` and
    assigns `np.nan` if the impact is not given.

    Note that ``anomaly_df`` can contain the anomaly information for many
    different timeseries. This is enabled by allowing multiple metrics and
    dimensions to be listed in the same anomaly dataframe. Columns can indicate
    the metric name and dimension value.

    This function first subsets the ``anomaly_df`` to the relevant rows for the
    ``value_col`` as specified by ``filter_by_dict``, then makes the specified
    adjustments to ``df``.

    Parameters
    ----------
    df : `pandas.DataFrame`
        A data frame which includes the timestamp column
        as well as the value column.
    time_col : `str`
        The column name in ``df`` representing time for the time series data.
        The time column can be anything that can be parsed by `pandas.DatetimeIndex`.
    value_col: `str`
        The column name which has the value of interest to be forecasted.
    anomaly_df : `pandas.DataFrame`
        A dataframe which includes the anomaly information for
        the input series (``df``) but potentially for multiple series and dimensions.

        This dataframe must include these two columns:

            - ``start_date_col``
            - ``end_date_col``

        and include

             - ``adjustment_delta_col`` if it is not None in the function call.

         Moreover if dimensions are requested by passing the
         ``filter_by_dict`` argument (not None), all of this dictionary
         keys must also appear in ``anomaly_df``.

        Here is an example::

            anomaly_df = pd.DataFrame({
                "start_date": ["1/1/2018", "1/4/2018", "1/8/2018", "1/10/2018"],
                "end_date": ["1/2/2018", "1/6/2018", "1/9/2018", "1/10/2018"],
                "adjustment_delta": [np.nan, 3, -5, np.nan],
                # extra columns for filtering
                "metric": ["y", "y", "z", "z"],
                "platform": ["MOBILE", "MOBILE", "DESKTOP", "DESKTOP"],
                "vertical": ["ads", "sales", "ads", "ads"],
            })

        In the above example,

            - "start_date" is the start date of the anomaly, which is provided using the argument ``start_date_col``.
            - "end_date" is the end date of the anomaly, which is provided using the argument ``end_date_col``.
            - "adjustment_delta" is the column which includes the delta if it is known. The name of this
              column is provided using the argument ``adjustment_delta_col``. Use `numpy.nan` if the
              adjustment size is not known, and the adjusted value will be set to `numpy.nan`.
            - "metric", "platform", and "vertical" are example columns for filtering. They
              contain the metric name and dimensions for which the anomaly is applicable.
              ``filter_by_dict` is used to filter on these columns to get the relevant
              anomalies for the timeseries represented by ``df[value_col]``.

    start_date_col : `str`, default ``START_DATE_COL``
        The column name in ``anomaly_df`` representing the start timestamp of
        the anomalous period, inclusive.
        The format can be anything that can be parsed by pandas DatetimeIndex.
    end_date_col : `str`, default ``END_DATE_COL``
        The column name in anomaly_df representing the start timestamp of
        the anomalous period, inclusive.
        The format can be anything that can be parsed by pandas DatetimeIndex.
    adjustment_delta_col : `str` or None, default None
        The column name in ``anomaly_df`` for the impact delta of the anomalies
        on the values of the series.

        If the value is available, it will be used to adjust the timeseries values
        in the given period by adding or subtracting this value to the raw series
        values in that period. Whether to add or subtract is specified by
        ``adjustment_method``.
        If the value for a row is "" or np.nan, the adjusted value is set to np.nan.

        If ``adjustment_delta_col`` is None, all adjusted values are set to np.nan.
    filter_by_dict : `dict` [`str`, `any`] or None, default None
        A dictionary whose keys are column names of ``anomaly_df``,
        and values are the desired value for that column (e.g. a string or int).
        If the value is an iterable (list, tuple, set), then it enumerates
        all allowed values for that column.

        This dictionary is used to filter ``anomaly_df`` to the matching anomalies.
        This helps when the ``anomaly_df`` includes the anomalies for various metrics
        and dimensions, so matching is needed to get the relevant anomalies for ``df``.

        Columns in ``anomaly_df`` can contain information on metric name,
        metric dimension (e.g. mobile/desktop), issue severity, etc. for filtering.
    filter_by_value_col: `str` or None, default None
        If provided, ``{filter_by_value_col: value_col}`` is added to ``filter_by_dict``
        for filtering. This filters ``anomaly_df`` to rows where
        ``anomaly_df[filter_by_value_col] == value_col``.

        If ``value_col`` is the metric name, this is a convenient way to find anomalies
        matching the metric name.
    adjustment_method : `str` ("add" or "subtract"), default "add"
        How the adjustment in ``anomaly_df`` should be used to adjust
        the value in ``df``.

            - If "add", the value in ``adjustment_delta_col`` is added to the original value.
            - If "subtract", it is subtracted from the original value.

    Returns
    -------
    Result : `dict`
        A dictionary with the following items (specified by key):

        - "adjusted_df": `pandas.DataFrame`
            A dataframe identical to the input dataframe ``df``, but with
            ``value_col`` updated to the adjusted values.
        - "augmented_df": `pandas.DataFrame`
            A dataframe identical to the input dataframe ``df``, with
            one extra column for the adjusted values: ``f"adjusted_{value_col}"``.
            ``value_col`` retains the original values.
            This is useful to inspect which values have changed.

    """
    df = df.copy()
    augmented_df = df.copy()
    anomaly_df = anomaly_df.copy()
    new_value_col = f"adjusted_{value_col}"

    if new_value_col in df.columns:
        raise ValueError(
            f"`df` cannot include this column name: {new_value_col}."
            f"This is because {new_value_col} will be autogenerated to include"
            "adjusted values in `augmented_df`.")
    if adjustment_method not in ["add", "subtract"]:
        raise ValueError(f"`adjustment_method` '{adjustment_method}' is not recognized, "
                         f"must be one of ['add', 'subtract']")
    if adjustment_delta_col is not None and adjustment_method == "subtract":
        anomaly_df[adjustment_delta_col] *= -1.0

    if filter_by_value_col is not None:
        if filter_by_dict is not None:
            filter_by_dict[filter_by_value_col] = value_col
        else:
            filter_by_dict = {filter_by_value_col: value_col}
    # If `filter_by_dict` is passed, we subset the anomaly dataframe
    # to filter on additional column values as specified.
    # Allows columns in `anomaly_df` indicating the metric name and dimension value
    # to be used to filter the rows.
    if filter_by_dict is not None:
        relevant_rows = pd.Series(np.repeat(True, anomaly_df.shape[0]))
        for filter_col, allowed_value in filter_by_dict.items():
            if filter_col not in anomaly_df.columns:
                raise ValueError(
                    f"Column '{filter_col}' was requested by `filter_by_dict`, "
                    "but the corresponding column name in `anomaly_df` is not found.")
            if isinstance(allowed_value, (tuple, list, set)):
                relevant_rows &= anomaly_df[filter_col].isin(allowed_value)
            else:
                relevant_rows &= (anomaly_df[filter_col] == allowed_value)
        anomaly_df = anomaly_df[relevant_rows]

    # Adjusts the values of the series using the anomaly information given
    # in the i-th column of the ``anomaly_df`` (after all the subsetting is done).
    augmented_df[new_value_col] = augmented_df[value_col]
    try:
        time_values = pd.to_datetime(augmented_df[time_col])
        anomaly_df[start_date_col] = pd.to_datetime(anomaly_df[start_date_col])
        anomaly_df[end_date_col] = pd.to_datetime(anomaly_df[end_date_col])
    except Exception as e:
        warnings.warn(f"Dates could not be parsed by `pandas.to_datetime`, using string comparison "
                      f"for dates instead. Error message:\n{e}")
        time_values = augmented_df[time_col].astype(str)
        anomaly_df[start_date_col] = anomaly_df[start_date_col].astype(str)
        anomaly_df[end_date_col] = anomaly_df[end_date_col].astype(str)
    for i in range(anomaly_df.shape[0]):
        row = anomaly_df.iloc[i]
        t1 = row[start_date_col]
        t2 = row[end_date_col]
        if t2 < t1:
            raise ValueError(
                f"End Time: {t2} cannot be before Start Time: {t1}, in ``anomaly_df``.")
        bool_index = (time_values >= t1) & (time_values <= t2)

        if adjustment_delta_col is not None:
            delta = row[adjustment_delta_col]
            if (delta != "") and not np.isnan(delta):
                augmented_df.loc[bool_index, new_value_col] = (
                        augmented_df.loc[bool_index, new_value_col] + delta)
            else:
                augmented_df.loc[bool_index, new_value_col] = np.nan
        else:
            augmented_df.loc[bool_index, new_value_col] = np.nan

    # Rewrites `df` values with the adjusted values
    df[value_col] = augmented_df[new_value_col]

    return {
        "adjusted_df": df,
        "augmented_df": augmented_df}



# BSD 2-CLAUSE LICENSE

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:

# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# #ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# original author: Reza Hosseini
"""Plotting functions in matplotlib."""

import colorsys

import matplotlib
import numpy as np
from matplotlib import pyplot as plt
from pandas.plotting import register_matplotlib_converters


matplotlib.use("agg")  # noqa: E402

register_matplotlib_converters()


def plt_compare_timeseries(
        df_dict,
        time_col,
        value_col,
        legends_dict=None,
        colors_dict=None,
        start_time=None,
        end_time=None,
        transform=lambda x: x,
        transform_name="",
        plt_title="",
        alpha=0.6,
        linewidth=4):
    """Compare a collection by of timeseries (given in `df_dict`)
       by overlaying them in the specified period between
       ``start_time`` and ``end_time``.

    :param df_dict: Dict[str, pd.DataFrame]
        The keys are the arbitrary labels for each dataframe provided by the user.
        The values are dataframes each containing a timeseries
        with `time_col` and `value_col` as columns.
    :param time_col: str
        The column denoting time in datetime format.
    :param value_col: str
        The value column of interest for the y axis.
    :param legends_dict: Optional[Dict[str, str]]
        Labels for plot legend.
        The keys are the df_dict labels (or a subset of them).
        The values are the labels appearing in the plot legend.
        If not provided or None, the `df_dict` keys will be used.
    :param colors_dict: Optional[Dict[str, str]]
        A dictionary determining the color for each series in `df_dict`.
        The keys are the df_dict labels (or a subset of them).
        The values are the colors appearing for each curve in the plot.
        If not provided or None, the colors will be generated.
    :param start_time: Optional[datetime.datetime]
        The start time of the series plot.
    :param end_time: Optional[datetime.datetime]
        The end time of the series plot.
    :param transform: Optional[func]
        A function to transform the y values before plotting.
    :param transform_name: Optional[str]
        The name of the transformation for using in the title.
    :param plt_title: str
        Plot title.
    :param alpha: Optional[float]
        Transparency of the curves.
    :param linewidth: Optional[float]
        The width of the curves.
    """
    if start_time is not None:
        for label, df in df_dict.items():
            df_dict[label] = df[df[time_col] >= start_time]

    if end_time is not None:
        for label, df in df_dict.items():
            df_dict[label] = df[df[time_col] <= end_time]

    n = len(df_dict)
    labels = list(df_dict.keys())
    if colors_dict is None:
        hsv_tuples = [(x * 1.0 / n, 0.5, 0.5) for x in range(n)]
        rgb_tuples = [(lambda x: colorsys.hsv_to_rgb(*x))(x) for x in hsv_tuples]
        colors_dict = {labels[i]: rgb_tuples[i] for i in range(n)}

    if legends_dict is None:
        legends_dict = {label: label for label in labels}

    fig, ax = plt.subplots()
    legend_patches = []
    for label in labels:
        df = df_dict[label]
        color = colors_dict[label]
        legend = legends_dict.get(label)
        # Avoids the following issue:
        # "ValueError: view limit minimum -36864.55 is less than 1 and is an invalid
        # Matplotlib date value. This often happens if you pass a non-datetime value
        # to an axis that has datetime units".
        dates = df[time_col].astype("O")
        # we add a legend for the given series if legend is not None
        # for that series
        if legend is not None:
            ax.plot(
                dates,
                transform(df[value_col]),
                alpha=alpha,
                color=color,
                label=legend,
                linewidth=linewidth)
            patch = matplotlib.patches.Patch(color=color, label=legend)
            legend_patches.append(patch)
        else:
            ax.plot(
                dates,
                transform(df[value_col]),
                alpha=alpha,
                color=color,
                linewidth=linewidth)

    legends = list(legends_dict.values())
    ax.legend(
        labels=[legend for legend in legends if legend is not None],
        handles=legend_patches)
    fig.autofmt_xdate()  # rotates the dates
    ax.set_title(plt_title + " " + transform_name)
    plt.show()


def plt_overlay_long_df(
        df,
        x_col,
        y_col,
        split_col,
        agg_dict=None,
        agg_col_names=None,
        overlay_color="black",
        agg_col_colors=None,
        plt_title=None):
    """Overlay by splitting wrt values of a column (split_col).
        If some agg metrics (specified by agg_dict) are also desired,
        we overlay the aggregate metrics as well.
    :param df: pd.DataFrame
        data frame which includes the data
    :param x_col: str
        the column for the values for the x-axis
    :param y_col: str
        the column for the values for the y-axis
    :param split_col: str
        the column which is used to split the data to various
        parts to be overlayed
    :param agg_dict: Optional[dict]
        a dictionary to specify aggregations.
        we could calculate multiple metrics for example mean and median
    :param agg_col_names: optional[list[str]]
        names for the aggregated columns.
        if not provided it will be generated during aggregations
    :param overlay_color: Optional[str]
        the color of the overlayed curves.
        The color will be transparent so we can see the curves overlap
    :param agg_col_colors: Optional[str]
        the color of the aggregate curves
    :param plt_title: Optional[str]
        the title of the plot.
        It will default to y_col if not provided
    """
    g = df.groupby([split_col], as_index=False)
    df_num = len(g)
    alpha = 5.0 / df_num

    for name, df0 in g:
        plt.plot(
            df0[x_col],
            df0[y_col],
            color=overlay_color,
            alpha=alpha,
            label=None)

    agg_df = None
    if agg_dict is not None:
        g2 = df.groupby([x_col], as_index=False)
        agg_df = g2.agg(agg_dict)
        agg_df.columns = [" ".join(col).strip() for col in agg_df.columns.values]
        if agg_col_names is not None:
            agg_df.columns = [x_col] + agg_col_names

        if agg_col_colors is None:
            n = len(agg_col_names)
            hsv_tuples = [(x * 1.0 / n, 0.5, 0.5) for x in range(n)]
            rgb_tuples = [colorsys.hsv_to_rgb(*x) for x in hsv_tuples]
            agg_col_colors = rgb_tuples

        legend_patches = []

        for i in range(len(agg_col_names)):
            col = agg_col_names[i]
            color = agg_col_colors[i]
            plt.plot(agg_df[x_col], agg_df[col], label=col, color=color)
            patch = matplotlib.patches.Patch(color=color, label=col)
            legend_patches.append(patch)

        plt.legend(labels=agg_col_names, handles=legend_patches)

    if plt_title is None:
        plt_title = y_col

    plt.title(plt_title)
    plt.xlabel(x_col)
    plt.ylabel(y_col)

    return agg_df


def plt_overlay_with_bands(
        df,
        x_col,
        y_col,
        split_col,
        perc=(25, 75),
        overlay_color="black",
        agg_col_colors=("blue", "black", "red"),
        plt_title=None):
    """Overlay by splitting wrt a column and plot wrt time.
        We also add quantile (percentile) bands
    :param df: pd.DataFrame
        data frame which includes the data
    :param x_col: str
        the column for the values for the x-axis
    :param y_col: str
        the column for the values for the y-axis
    :param split_col: str
        the column which is used to split the data
        to various partitions t be overlayed
    :param perc: tuple[float, float]
        the percentiles for the bands.
        The default is 25 and 75 percentiles
    :param overlay_color: Optional[str]
        the color of the overlayed curves.
        The color will be transparanet so we can see the curves overlap
    :param agg_col_colors: Optional[str]
        the color of the aggregate curves
    :param plt_title: Optional[str]
        the title of the plot.
        It will default to y_col if not provided
    """

    def quantile_fcn(q):
        return lambda x: np.nanpercentile(a=x, q=q)

    agg_dict = {
        y_col: [np.nanmean, quantile_fcn(perc[0]), quantile_fcn(perc[1])]
    }

    res = plt_overlay_long_df(
        df=df,
        x_col=x_col,
        y_col=y_col,
        split_col=split_col,
        agg_dict=agg_dict,
        agg_col_names=["mean", ("Q" + str(perc[0])), ("Q" + str(perc[1]))],
        overlay_color=overlay_color,
        agg_col_colors=agg_col_colors,
        plt_title=plt_title)

    return res


def plt_longterm_ts_agg(
        df,
        time_col,
        window_col,
        value_col,
        color_col=None,
        agg_func=np.nanmean,
        plt_title="",
        color="blue",
        choose_color_func=None):
    """Make a longterm avg plot by taking the average in a window
        and moving that across the time.
        That window can be a week, month, year etc

    :param df: pd.DataFrame
        The data frame with the data
    :param time_col: str
        The column with timestamps
    :param window_col: str
        This is a column which represents a coarse time granularity
        eg "2018-01" represents year-month
        In this case the avg across that month is calculated
    :param value_col: str
        The column denoting the values
    :param color_col: Optional[str]
        The column denoting the color for each point.
        We allow the curve color to change.
        When aggregating for each window the first color
        appearing in that window will be used by default.
    :param agg_func: Optional[func]
        the aggregation function to be used across the window
    :param plt_title: Optional[str]
        Plot title
    :param color: Optional[str]
        Color of the curve if it is not provided by `color_col`
        in original dataframe
    :param choose_color_func: func
        Function to determine how pointwise colors are translated to aggregated
        to be used for the aggregate curve.
        The default choosed the first color appearing in the data for
        the corresponding slice.
    :return pd.DataFrame
        The aggregated data frame constructed for the plot
    """
    df = df.copy()

    # this is for choosing the color for the aggregated curve
    # after averaging
    if choose_color_func is None:
        def choose_color_func(color_values):
            return color_values.iloc[0]

    if color_col is None:
        color_col = "curve_color"
        df["curve_color"] = color

    agg_dict = {
        time_col: np.nanmean,
        value_col: agg_func,
        color_col: choose_color_func}

    g = df.groupby([window_col], as_index=False)
    df_agg = g.agg(agg_dict)

    df_agg.columns = [window_col, time_col, value_col, color_col]
    df_agg.sort_values(by=time_col, inplace=True)
    # plotting the whole curve in grey
    plt.plot(
        df_agg[time_col],
        df_agg[value_col],
        alpha=0.5,
        color="grey")
    plt.xlabel(time_col)
    plt.ylabel(value_col)
    colors_set = set(list(df[color_col].values))
    # adding scatter points in colors for each part of curve
    # based on `color_col`
    for c in colors_set:
        df_agg_subset = df_agg.loc[
            df_agg[color_col] == c].reset_index(drop=True)
        plt.scatter(
            df_agg_subset[time_col].values,
            df_agg_subset[value_col].values,
            alpha=0.5,
            color=c)
    plt.title(plt_title)

    return df_agg



# BSD 2-CLAUSE LICENSE

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:

# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# #ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# original author: Reza Hosseini
"""Utility functions for test cases."""

import datetime
import warnings

import numpy as np
import pandas as pd

from greykite.common.constants import ADJUSTMENT_DELTA_COL
from greykite.common.constants import END_DATE_COL
from greykite.common.constants import EVENT_DF_LABEL_COL
from greykite.common.constants import METRIC_COL
from greykite.common.constants import START_DATE_COL
from greykite.common.constants import TIME_COL
from greykite.common.constants import VALUE_COL
from greykite.common.features.timeseries_features import add_daily_events
from greykite.common.features.timeseries_features import build_time_features_df
from greykite.common.features.timeseries_features import fourier_series_multi_fcn
from greykite.common.features.timeseries_features import get_default_origin_for_time_vars
from greykite.common.features.timeseries_features import get_fourier_col_name
from greykite.common.features.timeseries_features import get_holidays
from greykite.common.python_utils import assert_equal


def generate_df_for_tests(
        freq,
        periods,
        train_start_date=datetime.datetime(2018, 7, 1),
        train_end_date=None,
        train_frac=0.8,
        conti_year_origin=None,
        noise_std=2.0,
        remove_extra_cols=True,
        autoreg_coefs=None,
        fs_coefs=[-1, 3, 4],
        growth_coef=3.0,
        growth_pow=1.1,
        intercept=0.0):
    """Generates dataset for unit tests.

    :param freq: str
        pd.date_range freq parameter, e.g. H or D
    :param periods: int
        number of periods to generate
    :param train_start_date: datetime.datetime
        train start date
    :param train_end_date: Optional[datetime.datetime]
        train end date
    :param train_frac: Optional[float]
        fraction of data to use for training
        only used if train_end_date isn't provided
    :param noise_std: float
        standard deviation of gaussian noise
    :param conti_year_origin: float
        the time origin for continuous time variables
    :param remove_extra_cols: bool
        whether to remove extra columns besides TIME_COL, VALUE_COL
    :param autoreg_coefs: Optional[List[int]]
        The coefficients for the autoregressive terms.
        If provided the generated series denoted mathematically by Y(t) will be
        converted as follows:
        Y(t) -> Y(t) + c1 Y(t-1) + c2 Y(t-2) + c3 Y(t-3) + ...
        where autoreg_coefs = [c1, c2, c3, ...]
        In this fashion, the obtained series will have autoregressive
        properties not explained by seasonality and growth.
    :param fs_coefs: List[float]
        The fourier series coefficients used.
    :param growth_coef: float
        Multiplier for growth
    :param growth_pow: float
        Power for growth, as function of continuous time
    :param intercept: float
        Constant term added to Y(t)

    :return: Dict[str, any]
        contains full dataframe, train dataframe, test dataframe,
        and nrows in test dataframe
    """
    np.random.seed(123)

    date_list = pd.date_range(
        start=train_start_date,
        periods=periods,
        freq=freq).tolist()

    df0 = pd.DataFrame({TIME_COL: date_list})
    if conti_year_origin is None:
        conti_year_origin = get_default_origin_for_time_vars(df0, TIME_COL)
    time_df = build_time_features_df(
        dt=df0[TIME_COL],
        conti_year_origin=conti_year_origin)
    df = pd.concat([df0, time_df], axis=1)
    df["growth"] = growth_coef * (df["ct1"] ** growth_pow)

    func = fourier_series_multi_fcn(
        col_names=["toy", "tow", "tod"],
        periods=[1.0, 7.0, 24.0],
        orders=[1, 1, 1],
        seas_names=None)

    res = func(df)
    df_seas = res["df"]
    df = pd.concat([df, df_seas], axis=1)

    df[VALUE_COL] = (
            intercept
            + df["growth"]
            + fs_coefs[0] * df[get_fourier_col_name(1, "tod", function_name="sin")]
            + fs_coefs[1] * df[get_fourier_col_name(1, "tow", function_name="sin")]
            + fs_coefs[2] * df[get_fourier_col_name(1, "toy", function_name="sin")]
            + noise_std * np.random.normal(size=df.shape[0]))

    if autoreg_coefs is not None:
        df["temporary_new_value"] = df[VALUE_COL]
        k = len(autoreg_coefs)
        for i in range(k):
            df["temporary_new_value"] = (
                df["temporary_new_value"] +
                autoreg_coefs[i]*df[VALUE_COL].shift(-i)).bfill()
        df[VALUE_COL] = df["temporary_new_value"]
        del df["temporary_new_value"]

    if train_end_date is None:
        train_rows = np.floor(train_frac * df.shape[0]).astype(int)
        train_end_date = df[TIME_COL][train_rows]

    if remove_extra_cols:
        df = df[[TIME_COL, VALUE_COL]]
    train_df = df.loc[df[TIME_COL] <= train_end_date]
    test_df = df.loc[df[TIME_COL] > train_end_date]
    fut_time_num = test_df.shape[0]

    return {
        "df": df,
        "train_df": train_df.reset_index(drop=True),
        "test_df": test_df.reset_index(drop=True),
        "fut_time_num": fut_time_num,
    }


def generate_df_with_holidays(freq, periods):
    # generate data
    df = generate_df_for_tests(freq, periods, remove_extra_cols=False)["df"]

    # generate holidays
    countries = ["US", "India"]
    event_df_dict = get_holidays(countries, year_start=2015, year_end=2025)

    for country in countries:
        event_df_dict[country][EVENT_DF_LABEL_COL] = country + "_holiday"

    df = add_daily_events(
        df=df,
        event_df_dict=event_df_dict,
        date_col=TIME_COL,
        regular_day_label="")

    df[VALUE_COL] = (
            df[VALUE_COL]
            + 2 * (df["events_US"] == "US_holiday") * df[get_fourier_col_name(1, "tod", function_name="sin")]
            + 3 * (df["events_US"] == "US_holiday") * df[get_fourier_col_name(1, "tod", function_name="cos")]
            + 4 * (df["events_India"] == "India_holiday") * df[get_fourier_col_name(1, "tod", function_name="cos")])

    df = df[[TIME_COL, VALUE_COL]]
    thresh = datetime.datetime(2019, 8, 1)
    train_df = df[df[TIME_COL] <= thresh]
    test_df = df[df[TIME_COL] > thresh]
    fut_time_num = test_df.shape[0]

    return {
        "df": df,
        "train_df": train_df,
        "test_df": test_df,
        "fut_time_num": fut_time_num,
    }


def generate_df_with_reg_for_tests(
        freq,
        periods,
        train_start_date=datetime.datetime(2018, 7, 1),
        train_end_date=None,
        train_frac=0.8,
        conti_year_origin=None,
        noise_std=2.0,
        remove_extra_cols=True,
        mask_test_actuals=False):
    """Generates dataset for unit tests that includes regressor columns
    :param freq: str
        pd.date_range freq parameter, e.g. H or D
    :param periods: int
        number of periods to generate
    :param train_start_date: datetime.datetime
        train start date
    :param train_end_date: Optional[datetime.datetime]
        train end date
    :param train_frac: Optional[float]
        fraction of data to use for training
        only used if train_end_date isn't provided
    :param noise_std: float
        standard deviation of gaussian noise
    :param conti_year_origin: float
        the time origin for continuous time variables
    :param remove_extra_cols: bool
        whether to remove extra columns besides TIME_COL, VALUE_COL
    :param mask_test_actuals: bool
        whether to set y values to np.NaN in the test set.
    :return: Dict with train dataframe, test dataframe, and nrows in test dataframe
    """
    np.random.seed(123)

    result_list = generate_df_for_tests(
        freq,
        periods,
        train_start_date=train_start_date,
        train_end_date=train_end_date,
        train_frac=train_frac,
        conti_year_origin=conti_year_origin,
        noise_std=noise_std,
        remove_extra_cols=False)

    df = result_list["df"]
    df["regressor1"] = (
            df["growth"]
            + 4 * df[get_fourier_col_name(1, "tow", function_name="sin")]
            - 3 * df[get_fourier_col_name(1, "tod", function_name="sin")]
            + 7 * df[get_fourier_col_name(1, "toy", function_name="sin")]
            - noise_std * np.random.normal(size=df.shape[0]))

    df["regressor2"] = (
            df["growth"]
            + 1 * df[get_fourier_col_name(1, "tow", function_name="sin")]
            - 2 * df[get_fourier_col_name(1, "tod", function_name="sin")]
            + 3 * df[get_fourier_col_name(1, "toy", function_name="sin")]
            + noise_std * np.random.normal(size=df.shape[0]))

    df["regressor3"] = (
            df["growth"]
            + 9 * df[get_fourier_col_name(1, "tow", function_name="sin")]
            - 8 * df[get_fourier_col_name(1, "tod", function_name="sin")]
            + 5 * df[get_fourier_col_name(1, "toy", function_name="sin")]
            + noise_std * np.random.normal(size=df.shape[0]))

    df["regressor_bool"] = np.random.rand(df.shape[0]) > 0.3
    df["regressor_categ"] = np.random.choice(a=["c1", "c2", "c3"], size=df.shape[0], p=[0.1, 0.2, 0.7])

    if train_end_date is None:
        train_rows = np.floor(train_frac * df.shape[0]).astype(int)
        train_end_date = df[TIME_COL][train_rows]

    regressor_cols = [
        "regressor1",
        "regressor2",
        "regressor3",
        "regressor_bool",
        "regressor_categ"]
    if remove_extra_cols:
        df = df[[TIME_COL, VALUE_COL] + regressor_cols]

    if mask_test_actuals:
        # False positive warning:
        #   pandas.core.common.SettingWithCopyError:
        #   A value is trying to be set on a copy of a slice from a DataFrame.
        #   Try using .loc[row_indexer,col_indexer] = value instead
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            df.loc[df[TIME_COL] > train_end_date, VALUE_COL] = np.NaN

    train_df = df.loc[df[TIME_COL] <= train_end_date]
    test_df = df.loc[df[TIME_COL] > train_end_date]
    fut_time_num = test_df.shape[0]

    return {
        "df": df,
        "train_df": train_df,
        "test_df": test_df,
        "fut_time_num": fut_time_num,
        "regressor_cols": regressor_cols,
    }


def daily_data_reg():
    return generate_df_with_reg_for_tests(
        freq="D",
        periods=500,
        train_start_date=datetime.datetime(2018, 1, 1),
        conti_year_origin=2018)


def hourly_data_reg():
    """Generate 500 days of hourly data for tests"""
    return generate_df_with_reg_for_tests(
        freq="H",
        periods=24*500,
        train_start_date=datetime.datetime(2018, 1, 1),
        conti_year_origin=2018)


def generate_test_changepoint_df(
        freq="D",
        periods=200,
        n_changepoints=3,
        signal_strength=1/5,
        err_std=1.0):
    """Generates df to test change points

    The generated df is simple zigzag shaped over time with noise,
    with user specified frequency, length and number of change points.

    Parameters
    ----------
    freq : `DateOffset`, `Timedelta` or `str`, default is "D"
        The data frequency.
    periods : `int`, default is 200
        How many observation in frequency ``freq`` to be generated.
    n_changepoints : `int`
        How many change points to be included.
    signal_strength : `float`
        Time series signal strength.
    err_std : `float`
        Standard deviation of error.

    Returns
    -------
    df : `pandas.DataFrame`
        The generated df with columns:
            `"ts"` : time column.
            `"y"` : value column.
    """
    np.random.seed(123)
    start = "2020-01-01"
    y = np.array([0])
    nochange_period = np.ceil(periods / (n_changepoints + 1))
    for i in range(n_changepoints + 1):
        temp = np.arange(nochange_period + 1)
        if i % 2 == 1:
            temp = temp[::-1]
        temp = temp[1:]  # remove the repeated term
        y = np.concatenate([y, temp])
    y = y * signal_strength + np.random.randn(len(y)) * err_std
    df = pd.DataFrame(
        {
            "ts": pd.date_range(start=start, freq=freq, periods=periods),
            "y": y[:periods]
        }
    )
    return df


def generate_anomalous_data(periods=10):
    ts = pd.date_range(start="1/1/2018", periods=periods)
    df = pd.DataFrame({
        "ts": ts,
        "y": range(periods),
        "z": range(20, 20+periods)})

    anomaly_df = pd.DataFrame({
        METRIC_COL: ["y", "y", "z", "z"],
        "platform": ["MOBILE", "MOBILE", "DESKTOP", "DESKTOP"],
        "vertical": ["ads", "sales", "ads", "ads"],
        START_DATE_COL: ["1/1/2018", "1/4/2018", "1/8/2018", "1/10/2018"],
        END_DATE_COL: ["1/2/2018", "1/6/2018", "1/9/2018", "1/10/2018"],
        ADJUSTMENT_DELTA_COL: [np.nan, 3., -5., np.nan]})

    for col in ["start_date", "end_date"]:
        anomaly_df[col] = pd.to_datetime(anomaly_df[col])

    return {
        "df": df,
        "anomaly_df": anomaly_df}


def generic_test_adjust_anomalous_data(
        value_col,
        adj_df_info,
        adj_values):
    """Generic test for the results of any given scenario"""
    augmented_df = adj_df_info["augmented_df"]
    adjusted_df = adj_df_info["adjusted_df"]
    assert list(adjusted_df.columns) == ["ts", "y", "z"]
    assert list(augmented_df.columns) == ["ts", "y", "z", f"adjusted_{value_col}"]
    assert_equal(
        adjusted_df[value_col],
        augmented_df[f"adjusted_{value_col}"],
        check_names=False)
    assert_equal(
        augmented_df[:len(adj_values)][f"adjusted_{value_col}"],
        adj_values,
        check_names=False)


def gen_sliced_df(
        sample_size_dict={"a": 100, "b": 200, "c": 300, "d": 8, "e": 3},
        seed_dict={"a": 301, "b": 167, "c": 593, "d": 893, "e": 191, "z": 397},
        err_magnitude_coef=1.0):
    """generates a data frame which includes a response variable with column name "y", and an
    estimate of y with column name "y_hat". y is designed to be a function of a categorical
    variable given in "x" and a continuous variable given in "z".
    For various levels of y, the variance of y is different and the sample size is also different.
    The noises added are all deterministic so that this data set does not change at random"""

    y_a = np.linspace(start=90, stop=100, num=sample_size_dict["a"], endpoint=True)
    np.random.seed(seed=seed_dict["a"])
    y_a_hat = y_a + err_magnitude_coef * np.random.random(len(y_a))

    y_b = np.linspace(start=190, stop=200, num=sample_size_dict["b"], endpoint=True)
    np.random.seed(seed=seed_dict["b"])
    y_b_hat = y_b + err_magnitude_coef * 1.5 * np.random.random(len(y_b))

    y_c = np.linspace(start=290, stop=300, num=sample_size_dict["c"], endpoint=True)
    np.random.seed(seed=seed_dict["c"])
    y_c_hat = y_c + err_magnitude_coef * 2 * np.random.random(len(y_c))

    y_d = np.linspace(start=490, stop=550, num=sample_size_dict["d"], endpoint=True)
    np.random.seed(seed=seed_dict["d"])
    y_d_hat = y_d + err_magnitude_coef * np.random.random(len(y_d))

    y_e = np.linspace(start=-10, stop=0, num=sample_size_dict["e"], endpoint=True)
    np.random.seed(seed=seed_dict["e"])
    y_e_hat = y_e + err_magnitude_coef * 3 * np.random.random(len(y_e))

    y = np.concatenate((y_a, y_b, y_c, y_d, y_e), axis=0, out=None)
    y_hat = np.concatenate((y_a_hat, y_b_hat, y_c_hat, y_d_hat, y_e_hat), axis=0, out=None)

    # also create a continuous covariate which might be useful in some simulations
    np.random.seed(seed=seed_dict["z"])
    z = np.random.random(len(y))
    y = y + 2 * z
    y_hat = y_hat + 2.1 * z

    df = pd.DataFrame({
        "x": ["a"]*len(y_a) + ["b"]*len(y_b) + ["c"]*len(y_c) + ["d"]*len(y_d) + ["e"]*len(y_e),
        "z": z,
        "y": y,
        "y_hat": y_hat})
    df["residual"] = df["y"] - df["y_hat"]
    # adding another categorical column
    df["z_categ"] = df["z"].round().astype(str)

    return df


def assert_eval_function_equal(f1, f2):
    """Checks whether evaluation functions return the same output for a few inputs"""
    y_true = pd.Series([3, 1, 3])
    y_pred = pd.Series([1, 4, 2])
    assert f1(y_true, y_pred) == f2(y_true, y_pred)

    y_true = pd.Series([-20, 2, -10, 3.0])
    y_pred = pd.Series(np.arange(4))
    assert f1(y_true, y_pred) == f2(y_true, y_pred)

    y_true = pd.Series([0])
    y_pred = pd.Series([0])
    assert f1(y_true, y_pred) == f2(y_true, y_pred)

    np.random.seed(92)
    y_true = np.random.random(42)
    y_pred = np.random.random(42)
    assert_equal(f1(y_true, y_pred), f2(y_true, y_pred))



# BSD 2-CLAUSE LICENSE

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:

# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# #ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# original author: Albert Chen, Reza Hosseini, Sayan Patra
"""sklearn estimator with common functionality between
SilverkiteEstimator and SimpleSilverkiteEstimator.
"""

from typing import Dict
from typing import Optional

import pandas as pd
from sklearn.exceptions import NotFittedError
from sklearn.metrics import mean_squared_error

from greykite.algo.common.col_name_utils import create_pred_category
from greykite.algo.forecast.silverkite.forecast_silverkite import SilverkiteForecast
from greykite.algo.forecast.silverkite.forecast_silverkite_helper import get_silverkite_uncertainty_dict
from greykite.algo.forecast.silverkite.silverkite_diagnostics import SilverkiteDiagnostics
from greykite.common import constants as cst
from greykite.common.time_properties import min_gap_in_seconds
from greykite.common.time_properties_forecast import get_simple_time_frequency_from_period
from greykite.sklearn.estimator.base_forecast_estimator import BaseForecastEstimator


class BaseSilverkiteEstimator(BaseForecastEstimator):
    """A base class for forecast estimators that fit using
    `~greykite.algo.forecast.silverkite.forecast_silverkite.SilverkiteForecast.forecast`.

    Notes
    -----
    Allows estimators that fit using
    `~greykite.algo.forecast.silverkite.forecast_silverkite.SilverkiteForecast.forecast`
    to share the same functions for input data validation,
    fit postprocessing, predict, summary, plot_components, etc.

    Subclasses should:

        - Implement their own ``__init__`` that uses a superset of the parameters here.
        - Implement their own ``fit``, with this sequence of steps:

            - calls ``super().fit``
            - calls ``SilverkiteForecast.forecast`` or ``SimpleSilverkiteForecast.forecast_simple`` and stores the result in ``self.model_dict``
            - calls ``super().finish_fit``

    Uses ``coverage`` to set prediction band width. Even though
    coverage is not needed by ``forecast_silverkite``, it is included
    in every ``BaseForecastEstimator`` to be used universally for
    forecast evaluation.

    Therefore, ``uncertainty_dict`` must be consistent with ``coverage``
    if provided as a dictionary. If ``uncertainty_dict`` is None or
    "auto", an appropriate default value is set, according to ``coverage``.

    Parameters
    ----------
    score_func : callable, optional, default mean_squared_error
        See `~greykite.sklearn.estimator.base_forecast_estimator.BaseForecastEstimator`.
    coverage : `float` between [0.0, 1.0] or None, optional
        See `~greykite.sklearn.estimator.base_forecast_estimator.BaseForecastEstimator`.
    null_model_params : `dict`, optional
        Dictionary with arguments to define DummyRegressor null model, default is `None`.
        See `~greykite.sklearn.estimator.base_forecast_estimator.BaseForecastEstimator`.
    uncertainty_dict : `dict` or `str` or None, optional
        How to fit the uncertainty model.
        See `~greykite.algo.forecast.silverkite.forecast_silverkite.SilverkiteForecast.forecast`.
        Note that this is allowed to be "auto". If None or "auto", will be set to
        a default value by ``coverage`` before calling ``forecast_silverkite``.

    Attributes
    ----------
    silverkite : Class or a derived class of `~greykite.algo.forecast.silverkite.forecast_silverkite.SilverkiteForecast`
        The silverkite algorithm instance used for forecasting
    silverkite_diagnostics : Class or a derived class of `~greykite.algo.forecast.silverkite.silverkite_diagnostics.SilverkiteDiagnostics`
        The silverkite class used for plotting and generating model summary.
    model_dict : `dict` or None
        A dict with fitted model and its attributes.
        The output of `~greykite.algo.forecast.silverkite.forecast_silverkite.SilverkiteForecast.forecast`.
    pred_cols : `list` [`str`] or None
        Names of the features used in the model.
    feature_cols : `list` [`str`] or None
        Column names of the patsy design matrix built by
        `~greykite.algo.common.ml_models.design_mat_from_formula`.
    df : `pandas.DataFrame` or None
        The training data used to fit the model.
    coef_ : `pandas.DataFrame` or None
        Estimated coefficient matrix for the model.
        Not available for ``random forest`` and ``gradient boosting`` methods and
        set to the default value `None`.
    _pred_category : `dict` or None
        A dictionary with keys being the predictor category and
        values being the predictors belonging to the category.
        For details, see
        `~greykite.sklearn.estimator.base_silverkite_estimator.BaseSilverkiteEstimator.pred_category`.
    extra_pred_cols : `list` or None
        User provided extra predictor names, for details, see
        `~greykite.sklearn.estimator.simple_silverkite_estimator.SimpleSilverkiteEstimator`
        or
        `~greykite.sklearn.estimator.silverkite_estimator.SilverkiteEstimator`.
    forecast : `pandas.DataFrame` or None
        Output of ``predict_silverkite``, set by ``self.predict``.
    model_summary : `class` or `None`
        The `~greykite.algo.common.model_summary.ModelSummary` class.

    See Also
    --------
    `~greykite.algo.forecast.silverkite.forecast_silverkite.SilverkiteForecast.forecast`
        Function performing the fit and predict.

    Notes
    -----
    The subclasses will pass ``fs_components_df`` to ``forecast_silverkite``. The model terms
    it creates internally are used to generate the component plots.

        - `~greykite.common.features.timeseries_features.fourier_series_multi_fcn` uses
          ``fs_components_df["names"]`` (e.g. ``tod``, ``tow``) to build the fourier series
          and to create column names.

        - ``fs_components_df["seas_names"]`` (e.g. ``daily``, ``weekly``) is appended
          to the column names, if provided.

    `~greykite.algo.forecast.silverkite.silverkite_diagnostics.SilverkiteDiagnostics.plot_silverkite_components` groups
    based on ``fs_components_df["seas_names"]`` passed to ``forecast_silverkite`` during fit.
    E.g. any column containing ``daily`` is added to daily seasonality effect. The reason
    is as follows:

        1. User can provide ``tow`` and ``str_dow`` for weekly seasonality.
        These should be aggregated, and we can do that only based on "seas_names".
        2. yearly and quarterly seasonality both use ``ct1`` as "names" column.
        Only way to distinguish those effects is via "seas_names".
        3. ``ct1`` is also used for growth. If it is interacted with seasonality, the columns become
        indistinguishable without "seas_names".

    Additionally, the function sets yaxis labels based on ``seas_names``:
    ``daily`` as ylabel is much more informative than ``tod`` as ylabel in component plots.
    """
    def __init__(
            self,
            silverkite: SilverkiteForecast = SilverkiteForecast(),
            silverkite_diagnostics: SilverkiteDiagnostics = SilverkiteDiagnostics(),
            score_func: callable = mean_squared_error,
            coverage: float = None,
            null_model_params: Optional[Dict] = None,
            uncertainty_dict: Optional[Dict] = None):
        # initializes null model
        super().__init__(
            score_func=score_func,
            coverage=coverage,
            null_model_params=null_model_params)

        # required in subclasses __init__
        self.uncertainty_dict = uncertainty_dict

        # set by `fit`
        # fitted model in dictionary format returned from
        # the `forecast_silverkite` function
        self.silverkite: SilverkiteForecast = silverkite
        self.silverkite_diagnostics: SilverkiteDiagnostics = silverkite_diagnostics
        self.model_dict = None
        self.pred_cols = None
        self.feature_cols = None
        self.df = None
        self.coef_ = None
        # Predictor category, lazy initialization as None.
        # Will be updated in property function pred_category when needed.
        self._pred_category = None
        self.extra_pred_cols = None  # all silverkite estimators should support this.

        # set by the predict method
        self.forecast = None
        # set by the summary method
        self.model_summary = None

    def __set_uncertainty_dict(self, X, time_col, value_col):
        """Checks if ``coverage`` is consistent with the ``uncertainty_dict``
        used to train the ``forecast_silverkite`` model. Sets ``uncertainty_dict``
        to a default value if ``coverage`` is provided, and vice versa.

        Parameters
        ----------
        X: `pandas.DataFrame`
            Input timeseries, with timestamp column,
            value column, and any additional regressors.
            The value column is the response, included in
            ``X`` to allow transformation by `sklearn.pipeline.Pipeline`.
        time_col: `str`
            Time column name in ``X``.
        value_col: `str`
            Value column name in ``X``.

        Notes
        -----
        Intended to be called by `fit`.

        ``X`` is necessary to define default parameters for
        ``uncertainty_dict`` if ``coverage`` is provided but ``uncertainty_dict is None``
         or ``uncertainty_dict=="auto"``.
        (NB: ``X`` would not be necessary and this function could called from __init__
        if ``forecast_silverkite`` provides a default value for ``uncertainty_dict``
        given the target coverage).
        """
        period = min_gap_in_seconds(df=X, time_col=time_col)
        simple_freq = get_simple_time_frequency_from_period(period).name

        # Updates `uncertainty_dict` if None or "auto" or missing quantiles,
        # to match ``coverage``.
        # Raises an exception if both are provided and they don't match.
        self.uncertainty_dict = get_silverkite_uncertainty_dict(
            uncertainty=self.uncertainty_dict,
            simple_freq=simple_freq,
            coverage=self.coverage)

        # Updates coverage if None, to match the widest interval of
        # ``uncertainty_dict``. If coverage is not None, they are
        # already consistent, but we set it anyway.
        if self.uncertainty_dict is not None:
            quantiles = self.uncertainty_dict["params"]["quantiles"]
            self.coverage = quantiles[-1] - quantiles[0]

    def fit(
            self,
            X,
            y=None,
            time_col=cst.TIME_COL,
            value_col=cst.VALUE_COL,
            **fit_params):
        """Pre-processing before fitting ``Silverkite`` forecast model.

        Parameters
        ----------
        X: `pandas.DataFrame`
            Input timeseries, with timestamp column,
            value column, and any additional regressors.
            The value column is the response, included in
            ``X`` to allow transformation by `sklearn.pipeline`.
        y: ignored
            The original timeseries values, ignored.
            (The ``y`` for fitting is included in ``X``).
        time_col: `str`
            Time column name in ``X``.
        value_col: `str`
            Value column name in ``X``.
        fit_params: `dict`
            additional parameters for null model.

        Notes
        -----
        Subclasses are expected to call this at the beginning of their ``fit`` method,
        before calling `~greykite.algo.forecast.silverkite.forecast_silverkite.SilverkiteForecast.forecast`.
        """
        # NB: calls `__set_uncertainty_dict` before `super().fit` to ensure
        # coverage is correct before fitting the null model.
        # (null model does not currently use `coverage`, but may in the future.)
        self.__set_uncertainty_dict(
            X=X,
            time_col=time_col,
            value_col=value_col)
        self.df = X

        super().fit(
            X=X,
            y=y,
            time_col=time_col,
            value_col=value_col,
            **fit_params)

    def finish_fit(self):
        """Makes important values of ``self.model_dict`` conveniently accessible.

        To be called by subclasses at the end of their ``fit`` method. Sets
        {``pred_cols``, ``feature_cols``, and ``coef_``}.
        """
        if self.model_dict is None:
            raise ValueError("Must set `self.model_dict` before calling this function.")

        self.pred_cols = self.model_dict["pred_cols"]
        self.feature_cols = self.model_dict["x_mat"].columns
        # model coefficients
        if hasattr(self.model_dict["ml_model"], "coef_"):
            self.coef_ = pd.DataFrame(
                self.model_dict["ml_model"].coef_,
                index=self.feature_cols)

        self._set_silverkite_diagnostics_params()
        return self

    def _set_silverkite_diagnostics_params(self):
        if self.silverkite_diagnostics is None:
            self.silverkite_diagnostics = SilverkiteDiagnostics()
        self.silverkite_diagnostics.set_params(self.pred_category, self.time_col_, self.value_col_)

    def predict(self, X, y=None):
        """Creates forecast for the dates specified in ``X``.

        Parameters
        ----------
        X: `pandas.DataFrame`
            Input timeseries with timestamp column and any additional regressors.
            Timestamps are the dates for prediction.
            Value column, if provided in ``X``, is ignored.
        y: ignored.

        Returns
        -------
        predictions: `pandas.DataFrame`
            Forecasted values for the dates in ``X``. Columns:

                - ``TIME_COL``: dates
                - ``PREDICTED_COL``: predictions
                - ``PREDICTED_LOWER_COL``: lower bound of predictions, optional
                - ``PREDICTED_UPPER_COL``: upper bound of predictions, optional
                - [other columns], optional

            ``PREDICTED_LOWER_COL`` and ``PREDICTED_UPPER_COL`` are present
            if ``self.coverage`` is not None.
        """
        # Returns the cached result if applicable
        cached_predictions = super().predict(X=X)
        if cached_predictions is not None:
            return cached_predictions

        if self.model_dict is None:
            raise NotFittedError("Call `fit` before calling `predict`.")
        if self.pred_cols is None:
            raise NotFittedError("Subclass must call `finish_fit` inside the `fit` method.")

        pred_df = self.silverkite.predict(
            fut_df=X,
            trained_model=self.model_dict,
            past_df=None,
            new_external_regressor_df=None)["fut_df"]  # regressors are included in X

        self.forecast = pred_df
        # renames columns to standardized schema
        output_columns = {
            self.time_col_: cst.TIME_COL,
            cst.VALUE_COL: cst.PREDICTED_COL}

        # Checks if uncertainty is also returned.
        # If so, extract the upper and lower limits of the tuples in
        # ``uncertainty_col`` to be lower and upper limits of the prediction interval.
        # Note that the tuple might have more than two elements if more than two
        # ``quantiles`` are passed in ``uncertainty_dict``.
        uncertainty_col = f"{cst.VALUE_COL}_quantile_summary"
        if uncertainty_col in list(pred_df.columns):
            pred_df[cst.PREDICTED_LOWER_COL] = pred_df[uncertainty_col].apply(
                lambda x: x[0])
            pred_df[cst.PREDICTED_UPPER_COL] = pred_df[uncertainty_col].apply(
                lambda x: x[-1])
            output_columns.update({
                cst.PREDICTED_LOWER_COL: cst.PREDICTED_LOWER_COL,
                cst.PREDICTED_UPPER_COL: cst.PREDICTED_UPPER_COL,
                uncertainty_col: uncertainty_col})

        predictions = (pred_df[output_columns.keys()]
                       .rename(output_columns, axis=1))
        # Caches the predictions
        self.cached_predictions_ = predictions
        return predictions

    @property
    def pred_category(self):
        """A dictionary that stores the predictor names in each category.

        This property is not initialized until used. This speeds up the
        fitting process. The categories includes

            - "intercept" : the intercept.
            - "time_features" : the predictors that include
              `~greykite.common.constants.TIME_FEATURES`
              but not
              `~greykite.common.constants.SEASONALITY_REGEX`.
            - "event_features" : the predictors that include
              `~greykite.common.constants.EVENT_PREFIX`.
            - "trend_features" : the predictors that include
              `~greykite.common.constants.TREND_REGEX`
              but not
              `~greykite.common.constants.SEASONALITY_REGEX`.
            - "seasonality_features" : the predictors that include
              `~greykite.common.constants.SEASONALITY_REGEX`.
            - "lag_features" : the predictors that include
              `~greykite.common.constants.LAG_REGEX`.
            - "regressor_features" : external regressors and other predictors
              manually passed to ``extra_pred_cols``, but not in the categories above.
            - "interaction_features" : the predictors that include
              interaction terms, i.e., including a colon.

        Note that each predictor falls into at least one category.
        Some "time_features" may also be "trend_features".
        Predictors with an interaction are classified into all categories matched by
        the interaction components. Thus, "interaction_features" are already included
        in the other categories.
        """
        if self.model_dict is None:
            raise NotFittedError("Must fit before getting predictor category.")
        if self._pred_category is None:
            # extra_pred_cols could be None/list
            extra_pred_cols = [] if self.extra_pred_cols is None else self.extra_pred_cols
            # regressor_cols could be non-exist/None/list
            # the if catches non-exist and None
            regressor_cols = [] if getattr(self, "regressor_cols", None) is None else getattr(self, "regressor_cols")
            self._pred_category = create_pred_category(
                pred_cols=self.model_dict["x_mat"].columns,
                # extra regressors are specified via "regressor_cols" in simple_silverkite_estimator
                extra_pred_cols=extra_pred_cols + regressor_cols)
        return self._pred_category

    def summary(self, max_colwidth=20):
        if self.silverkite_diagnostics is None:
            self._set_silverkite_diagnostics_params()
        return self.silverkite_diagnostics.summary(self.model_dict, max_colwidth)

    def plot_components(self, names=None, title=None):
        if self.model_dict is None:
            raise NotFittedError("Call `fit` before calling `plot_components`.")
        if self.silverkite_diagnostics is None:
            self._set_silverkite_diagnostics_params()
        return self.silverkite_diagnostics.plot_components(self.model_dict, names, title)

    def plot_trend(self, title=None):
        """Convenience function to plot the data and the trend component.

        Parameters
        ----------
        title: `str`, optional, default `None`
            Plot title.

        Returns
        -------
        fig: `plotly.graph_objs.Figure`
            Figure.
        """
        if title is None:
            title = "Trend plot"
        return self.plot_components(names=["trend"], title=title)

    def plot_seasonalities(self, title=None):
        """Convenience function to plot the data and the seasonality components.

        Parameters
        ----------
        title: `str`, optional, default `None`
            Plot title.

        Returns
        -------
        fig: `plotly.graph_objs.Figure`
            Figure.
        """
        if title is None:
            title = "Seasonality plot"
        seas_names = [
            "DAILY_SEASONALITY",
            "WEEKLY_SEASONALITY",
            "MONTHLY_SEASONALITY",
            "QUARTERLY_SEASONALITY",
            "YEARLY_SEASONALITY"]
        return self.plot_components(names=seas_names, title=title)

    def plot_trend_changepoint_detection(self, params=None):
        """Convenience function to plot the original trend changepoint detection results.

        Parameters
        ----------
        params : `dict` or `None`, default `None`
            The parameters in `~greykite.algo.changepoint.adalasso.changepoint_detector.ChangepointDetector.plot`.
            If set to `None`, all components will be plotted.

            Note: seasonality components plotting is not supported currently. ``plot`` parameter must be False.

        Returns
        -------
        fig : `plotly.graph_objs.Figure`
            Figure.
        """
        if params is None:
            params = dict(
                observation=True,
                observation_original=True,
                trend_estimate=True,
                trend_change=True,
                yearly_seasonality_estimate=True,
                adaptive_lasso_estimate=True,
                seasonality_change=False,  # currently for trend only
                seasonality_change_by_component=False,
                seasonality_estimate=False,
                plot=False)
        else:
            # currently for trend only
            params["seasonality_change"] = False
            params["seasonality_estimate"] = False
            # need to return the figure object
            params["plot"] = False
        return self.model_dict["changepoint_detector"].plot(**params)



# BSD 2-CLAUSE LICENSE

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:

# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# #ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# original author: Albert Chen
"""Modification of `sklearn.metrics.make_scorer` and
`sklearn.metrics.scorer._PredictScorer` for
sklearn estimators whose predict function return a
`pandas.DataFrame` with the predicted value.
"""

import warnings
from functools import partial

import numpy as np

from greykite.common.constants import PREDICTED_COL


def _cached_call(cache, estimator, method, *args, **kwargs):
    """Call estimator with method and args and kwargs.
    This code is private in scikit-learn 0.24, so it is copied here.
    """
    if cache is None:
        return getattr(estimator, method)(*args, **kwargs)

    try:
        return cache[method]
    except KeyError:
        result = getattr(estimator, method)(*args, **kwargs)
        cache[method] = result
        return result


class _BaseScorer:
    """This code is private in scikit-learn 0.24, so it is copied here."""
    def __init__(self, score_func, sign, kwargs):
        self._kwargs = kwargs
        self._score_func = score_func
        self._sign = sign

    @staticmethod
    def _check_pos_label(pos_label, classes):
        if pos_label not in list(classes):
            raise ValueError(
                f"pos_label={pos_label} is not a valid label: {classes}"
            )

    def _select_proba_binary(self, y_pred, classes):
        """Select the column of the positive label in `y_pred` when
        probabilities are provided.
        Parameters
        ----------
        y_pred : ndarray of shape (n_samples, n_classes)
            The prediction given by `predict_proba`.
        classes : ndarray of shape (n_classes,)
            The class labels for the estimator.
        Returns
        -------
        y_pred : ndarray of shape (n_samples,)
            Probability predictions of the positive class.
        """
        if y_pred.shape[1] == 2:
            pos_label = self._kwargs.get("pos_label", classes[1])
            self._check_pos_label(pos_label, classes)
            col_idx = np.flatnonzero(classes == pos_label)[0]
            return y_pred[:, col_idx]

        err_msg = (
            f"Got predict_proba of shape {y_pred.shape}, but need "
            f"classifier with two classes for {self._score_func.__name__} "
            f"scoring"
        )
        raise ValueError(err_msg)

    def __repr__(self):
        kwargs_string = "".join([", %s=%s" % (str(k), str(v))
                                 for k, v in self._kwargs.items()])
        return ("make_scorer(%s%s%s%s)"
                % (self._score_func.__name__,
                   "" if self._sign > 0 else ", greater_is_better=False",
                   self._factory_args(), kwargs_string))

    def __call__(self, estimator, X, y_true, sample_weight=None):
        """Evaluate predicted target values for X relative to y_true.
        Parameters
        ----------
        estimator : object
            Trained estimator to use for scoring. Must have a predict_proba
            method; the output of that is used to compute the score.
        X : {array-like, sparse matrix}
            Test data that will be fed to estimator.predict.
        y_true : array-like
            Gold standard target values for X.
        sample_weight : array-like of shape (n_samples,), default=None
            Sample weights.
        Returns
        -------
        score : float
            Score function applied to prediction of estimator on X.
        """
        return self._score(partial(_cached_call, None), estimator, X, y_true,
                           sample_weight=sample_weight)

    def _factory_args(self):
        """Return non-default make_scorer arguments for repr."""
        return ""


class _PredictScorerDF(_BaseScorer):
    def _score(self, method_caller, estimator, X, y_true, sample_weight=None):
        """Evaluates predicted target values for ``X`` relative to ``y_true``.

        Modified from `sklearn.metrics.scorer._PredictScorer` to work with
        estimators that return a `pandas.DataFrame` with multiple
        columns, one of which contains the predicted values (``PREDICTED_COL``).

        Parameters
        ----------
        method_caller : callable
            Returns predictions given an estimator, method name, and other
            arguments, potentially caching results.
        estimator : `BaseForecastEstimator`
            Trained estimator to use for scoring. Must have a `predict`
            method that returns a pandas.DataFrame with a column
            :const:`~greykite.common.constants.PREDICTED_COL`.
            This column is the predicted value used to compute the score.
        X : array-like or sparse matrix
            Test data that will be fed to estimator.predict.
        y_true : array-like
            Gold standard target values for X.
        sample_weight : array-like, optional (default=None)
            Sample weights.

        Returns
        -------
        score : float
            Score function applied to prediction of estimator on X.
        """
        y_pred = method_caller(estimator, "predict", X)[PREDICTED_COL]
        if sample_weight is not None:
            score = self._score_func(y_true, y_pred,
                                     sample_weight=sample_weight,
                                     **self._kwargs)
        else:
            score = self._score_func(y_true, y_pred,
                                     **self._kwargs)

        if score is None:
            # Set to np.nan if undefined
            # (e.g. if MAPE, correlation, R2, or R2_null_model_score has division by 0).
            # This function cannot not return None to avoid the following error in
            # sklearn grid search (sklearn/model_selection/_validation.py:618):
            #   `scoring must return a number, got None (<class 'NoneType'>) instead. (scorer=score)`
            # `np.nan` is also the default value for `error_score` in
            # `~sklearn.model_selection.RandomizedSearchCV`. This split score is set
            # to this value if fit fails.
            warnings.warn("Score is undefined for this split, setting to `np.nan`.")
            score = np.nan

        return self._sign * score


def make_scorer_df(score_func, greater_is_better=True, **kwargs):
    """Make a scorer from a performance metric or loss function.

    This factory function wraps scoring functions for use in GridSearchCV
    and cross_val_score. It takes a score function, such as ``accuracy_score``,
    ``mean_squared_error``, ``adjusted_rand_index`` or ``average_precision``
    and returns a callable that scores an estimator's output.

    Modified from `sklearn.metrics.make_scorer` to work with
    estimators whose predict function returns a `pandas.DataFrame`
    with multiple columns, one of which contains the predicted values.

    Enabled any standard `score_func` to be used in grid search on a
    `~greykite.sklearn.estimator.base_forecast_estimator.BaseForecastEstimator`.

    Parameters
    ----------
    score_func : callable,
        Score function (or loss function) with signature
        ``score_func(y, y_pred, **kwargs)``.

    greater_is_better : boolean, default=True
        Whether score_func is a score function (default), meaning high is good,
        or a loss function, meaning low is good. In the latter case, the
        scorer object will sign-flip the outcome of the score_func.

    **kwargs : additional arguments
        Additional parameters to be passed to score_func.

    Returns
    -------
    scorer : callable
        Callable object that returns a scalar score; greater is better.
    """
    sign = 1 if greater_is_better else -1
    cls = _PredictScorerDF
    return cls(score_func, sign, kwargs)
